
# library(profvis) ## this is for profiling
library(sp)
library(rgeos)
library(tuneR)
#library(refund)
library(Matrix)
library(wrassp)
library(stringr)
library(igraph)
library(fdaPDE)
library(abind)
library(foreach)
library(doParallel)
library(expm)
library(fields)
library(magrittr)
#### extract a sound (internal function)
#' @examples
#' sound <- extract_sound_(sound.dir, bnc.data$wavfile[1], bnc.data$begintime[1], bnc.data$endtime[1])
#' @section Details:
#' The output sound amplitude is renormalized
extract_sound_ <- function(sound.dir, wavfile, begin.time, end.time, normalize.level, normalize.len)
{

    ## old and slow
    #fullsound <- readWave(paste0(sound.dir, wavfile))
    #sound <- fullsound[ (begin.time*fullsound@samp.rate) : (end.time*fullsound@samp.rate)] 

    sound <- tuneR::readWave(paste0(sound.dir, wavfile), from=begin.time, to=end.time, units='seconds')

    #
#    sound <- tuneR::normalize(sound, as.character(sound@bit))

    if(!is.null(normalize.level))
    {
    sound <- Wave(floor((2^(sound@bit-1)-1)*pmin(1, pmax(-1, normalize.level*sound@left/(max(1, sqrt(stats::filter(sound@left^2, rep(normalize.len/sound@samp.rate, sound@samp.rate/normalize.len))), na.rm=TRUE)))) ), bit=sound@bit, samp.rate=sound@samp.rate)
    }



    return(sound)
}


#' play bnc row
bnc_play <- function(bnc.row, ...)
{
    tuneR::play(bnc_to_sound(bnc.row, ...))
}

#' play bnc row using mplayer
bnc_mplay <- function(bnc.row, sound.dir.=sound.dir, add.begin=0, add.end=0)
{
    command <- paste0('mplayer ', sound.dir., bnc.row$wavfile, " -ss ", bnc.row$begintime - add.begin, ' -endpos ', bnc.row$endtime + add.end - bnc.row$begintime + add.begin) 
    print(command)
    system( command ) 
}

#' get amplitude of sound portion
bnc_amplitude  <-  function(wordinfo, sound.dir.=sound.dir, add.begin=0, add.end=0)
{
    sound.dir=sound.dir. 
    wavfile=wordinfo$wavfile 
    begin.time=wordinfo$begintime - add.begin
    end.time=wordinfo$endtime + add.end
    sound <- tuneR::readWave(paste0(sound.dir, wavfile), from=begin.time, to=end.time, units='seconds')
    return(max(abs(sound@left))/(2^15))
}

    
#### wrapper function to extract a sound
bnc_to_sound  <-  function(wordinfo, sound.dir.=sound.dir, output.file=NULL, add.begin=0, add.end=0, normalize.level=.8, normalize.len=10)
{
    sound <- do.call(extract_sound_, list(sound.dir=sound.dir., wavfile=wordinfo$wavfile, begin.time=wordinfo$begintime - add.begin, end.time=wordinfo$endtime + add.end, normalize.level=normalize.level, normalize.len=normalize.len))
    if(is.null(output.file))
    {
    return(invisible(sound))
    } else { ## output to file
        writeWave(sound, filename=output.file, extensible=FALSE)
    return(invisible(sound))
    }
}
# alias
extract_sound  <-  bnc_to_sound

#' extract longitude and latitude of the observations from  bnc row 
bnc_to_lonlat <- function(bnc.row)
{
    if(!exists('locations'))
    {
        load('locations.RData')
    }
    i <- which(as.character(locations$location.names) == bnc.row$placenamecleaned)
    if(length(i) > 0)
    {
        return(as.matrix(locations[i, 2:3]))
    } else {
        return(matrix(NA, 2, 1))
    }
}

#' extract formant information from bnc.row
bnc_to_pitch <- function(bnc.row, sound.dir.=sound.dir, add.begin=0, add.end=0, ...)
{
    output <- wrassp::mhsF0( paste0(sound.dir., bnc.row$wavfile), toFile=FALSE,  beginTime=bnc.row$begintime - add.begin, endTime=bnc.row$endtime + add.end, ...)$pitch
    output[output==0] <- NA
    return(output)
}

#' extract formant information from bnc.row
bnc_to_formant <- function(bnc.row, sound.dir.=sound.dir, add.begin=0, add.end=0, ...)
{
    output <- wrassp::forest( paste0(sound.dir., bnc.row$wavfile), toFile=FALSE,  beginTime=bnc.row$begintime - add.begin, endTime=bnc.row$endtime + add.end, ...)
    output$fm[output$fm==0] <- NA
    return(output)
}

#' extract pitch information from bnc.row
bnc_to_pitch_praat <- function(bnc.row, sound.dir.=sound.dir, add.begin=0, add.end=0, ...)
{
    praat.results.file.pitch <- tempfile() # tempfile for script
    script.praat <- tempfile(fileext='.praat') # tempfile for script

    cat(file=script.praat, sep='', 'namedata$ = "' , str_sub(bnc.row$wavfile, 1, -5) , '"\n',
        'dirdata$ = "', sound.dir., '"\n',
        "Open long sound file... 'dirdata$'/'namedata$'.wav
        select LongSound 'namedata$'
        Extract part... ", bnc.row$begin, " ", bnc.row$end, ' yes\n',
        "select Sound 'namedata$'
        Rename... tmpsound
        select Sound tmpsound
        To Pitch... 0.0 75 600
        select Pitch tmpsound
        To Matrix
        select Matrix tmpsound", "\n",
        'selectObject: "Matrix tmpsound"
        Save as headerless spreadsheet file: "', praat.results.file.pitch, '"'
        )

        print(paste(praat.soft, '--run', script.praat)) 
        system(paste(praat.soft, '--run', script.praat)) 

        pitch <- unlist(read.table(praat.results.file.pitch, na='0'))
        
        on.exit( file.remove( script.praat, praat.results.file.pitch) )
        return(invisible(pitch))
}

#' extract formant information from bnc.row
bnc_to_formant_praat <- function(bnc.row, sound.dir.=sound.dir, add.begin=0, add.end=0, ...)
{
    praat.results.file.formant <- tempfile() # tempfile for script
    script.praat <- tempfile(fileext='.praat') # tempfile for script

    cat(file=script.praat, sep='', 'namedata$ = "' , str_sub(bnc.row$wavfile, 1, -5) , '"\n',
        'dirdata$ = "', sound.dir., '"\n',
        "Open long sound file... 'dirdata$'/'namedata$'.wav
        select LongSound 'namedata$'
        Extract part... ", bnc.row$begin, " ", bnc.row$end, ' yes\n',
        "select Sound 'namedata$'
        #Edit
        Rename... tmpsound
        select Sound tmpsound
        To Formant (burg)... " , "0.001 5 5500 0.025 50", '\n',
        'select Formant tmpsound
        Down to Table: "no", "yes", 6, "yes", 3, "yes", 3, "yes"
        selectObject: "Table tmpsound"
        Save as comma-separated file: "', praat.results.file.formant, '"\n'
        )

        system(paste(praat.soft, '--run', script.praat)) 

        formants <- read.csv( praat.results.file.formant, header=TRUE, na='--undefined--', colClass=rep('numeric', 13))
        on.exit( file.remove( script.praat, praat.results.file.formant) )
        return(invisible(formants))
}


#' rescale time to 0, 1
rescale_time <- function(Y, len.out=100)
{
    Y <- as.matrix(Y)
    x.seq <- seq(0, 1, len=nrow(Y))
    apply(Y, 2, function(y)
          {
          approx(x.seq, y , seq(0, 1, len=len.out))$y
})
    #    approx.tmp <- approx(seq(0, 1, len=nrow(X)), X, seq(0, 1, len=len.out))$y
}


#a <- extract_sound(bnc.data[1, ])
## extract mfcc from a sound of class 'Wave'
sound_to_mfcc <- function(sound, ccord=39, lframe=80)
{
    wavfile <- tempfile('sound', fileext='.wav')
    writeWave(sound, filename=wavfile, extensible=FALSE)
    file_f0 <- tempfile('f0')
    file_fcc <- tempfile('fcc')
#    print(paste(ahocoder, wavfile,  file_f0,  file_fcc, paste0("--CCORD=", ccord) , paste0("--LFRAME=", lframe), sep=" "))
    ret <- system(
                  paste(ahocoder, wavfile,  file_f0,  file_fcc, paste0("--CCORD=", ccord) , paste0("--LFRAME=", lframe), sep=" ")
                  , intern=TRUE  )
    if(length(attr(ret, 'status')) != 0)
    { # if there is an error
        f0 <- NULL
        ret <- system( paste(ahocoder, wavfile,  file_f0,  file_fcc,
                             paste0("--CCORD=", ccord) , paste0("--LFRAME=",
                                                                lframe),
                             '--CCMETH=2',  sep=" "), intern=TRUE)
        if(length(attr(ret, 'status')) != 0)
        { # if there is still an error 
            return(invisible(list(mfcc=NULL, f0=NULL)))
        } 
    } else {
        ## reading f0
        to.read <- file(file_f0, "rb")
        f0=readBin(to.read, numeric(), size=4, n=1*2000)
        close(to.read)
    }
    ## reading mfcc
    to.read <- file(file_fcc, "rb")
    tmp.read=readBin(to.read, numeric(), size=4, n=(ccord+1)*2000)
    close(to.read)
    mfcc <- t(matrix(tmp.read, (ccord+1), length(tmp.read)/(ccord+1)))
    rm(tmp.read)

    return(invisible(mget(c('mfcc', 'f0'))))
}


## extract mfcc from a row of bnc.data
bnc_to_mfcc <- function(wordinfo, sound.dir.=sound.dir, ccord=39, lframe=80, ...)
{
    sound <- bnc_to_sound(wordinfo, sound.dir., output.file=NULL, ...)
    sound_to_mfcc(sound, ccord, lframe)
}
## alias 
extract_mfcc  <- bnc_to_mfcc 

## extract mfcc from a row of bnc.data using tuneR
bnc_to_mfcc_tuneR <- function(wordinfo, sound.dir.=sound.dir, ccord=39,  wintime=.025, hoptime=.001, ...)
{
    sound <- extract_sound(wordinfo, sound.dir., output.file=NULL)
    return(tuneR::melfcc(sound, wintime=wintime, hoptime=hoptime, numcep=ccord, ...))
}


##
mfcc_to_sound <- function(mfcc,  output.file=NULL, lframe=80)
{
    wavfile <- tempfile('sound', fileext='.wav')
    file_f0 <- tempfile('f0')
    file_fcc <- tempfile('fcc')

    zz <- file(file_f0, "wb")
    writeBin(mfcc$f0, zz, size = 4)
    close(zz)

    zz <- file(file_fcc, "wb")
    writeBin(as.vector(t(mfcc$mfcc)), zz, size = 4)
    close(zz)

    system(
           paste(ahodecoder, file_f0, file_fcc, wavfile, paste0("--LFRAME=", lframe),
                 sep=" ")
           )

    sound <- readWave(wavfile)
    if(is.null(output.file))
    {
    return(invisible(sound))
    } else { ## output to file
        writeWave(sound, filename=output.file, extensible=FALSE)
    return(invisible(sound))
    }
}

##
mfcc_play <- function(mfcc,  output.file=NULL, lframe=80)
{
    tuneR::play(mfcc_to_sound(mfcc, output.file, lframe))
}

## use 'tuneR::normalize' instead
#renormalize_sound <- function(sound, bit, amplitude=.8) {
#    return(round(sound/max(abs(sound))*(2^(bit - 1) - 1)*amplitude ))
#}

#### paste sounds together
paste_sounds_together <- function(wordinfos, sound.dir.=sound.dir, break.time=1, output.file=NULL)
{
    first.sound <- bnc_to_sound(wordinfos[1, ], sound.dir.)
    #first.sound  <- tuneR::normalize(first.sound, as.character(first.sound@bit))
    break.sequence <- rep(0, round(break.time*first.sound@samp.rate))
    other.sounds <- sapply(2:nrow(wordinfos), function(i)
                           {
           sound.tmp <- bnc_to_sound(wordinfos[i, ], sound.dir.)
           return(c(break.sequence, sound.tmp@left))
})
    
    sounds <- Wave(left=c(first.sound@left, unlist(other.sounds)), right=numeric(0), samp.rate=first.sound@samp.rate, bit=first.sound@bit)
    if(is.null(output.file))
    {
    return(invisible(sounds))
    } else { ## output to file
        writeWave(sounds, filename=output.file, extensible=FALSE)
        return(invisible(sounds))
    }
}




#### measure the distance between two (time scaled) MFCCs representations
#' @examples
#' d <- dist_mfcc(aligned.sounds[1, ,], aligned.sounds[2, ,])
#' @section Details:
#' The function compute the distance between two MFCC representation, weighting the coefficients with respect to their average absolute value.
dist_mfcc <- function(MFCC1, MFCC2)
{
  nt <- min(dim(MFCC1)[1], dim(MFCC2)[1])
  nc <- dim(MFCC1)[2]
  chdist <- rep(NA, nc)
  # for (k in 1:nc)
  #{
  #   chdist[k] <- acos(abs(cor(MFCC1[1:nt, k], MFCC2[1:nt, k])))
  # }
#   weights <- rep(NA, nc)
#   weights <- c(12:1, exp(-(1:(nc - 12))))
# #  for (k in 1:nc)
  #{
# #    weights[k] <- sum(abs(MFCC1[k, ]+sum(MFCC2[k, ]))/2)
# #  }
#   weights <- weights/sum(weights)
   for (k in 1:nc)
   {
     chdist[k] <- sum((MFCC1[, k]-MFCC2[, k])^2)*(1/nt)
   }
  #d <- sum(weights*chdist)
  d <- sqrt(sum(chdist))
  return(d)
}


#### compute the distance matrix between a set of MFCCs data
#' @examples
#' d <- dist_matrix(aligned.sounds)
#' @section Details:
#' 


dist_matrix <- function(MFCC.data)
{
  n <- dim(MFCC.data)[1]
  distance.matrix <- matrix(0, n, n)
  for (k in 2:n)
  {
    for (j in 1:(k-1))
    {
      distance.matrix[k, j] <- dist_mfcc(MFCC.data[k, ,], MFCC.data[j, ,])
      distance.matrix[j, k] <- distance.matrix[k, j]
    }
  }
  return(invisible(distance.matrix))
}



#### align the MFCC coefficients using the Fisher-Rao elastic warping on a chosen coepstral coefficient (default 1)
#' @examples
#' aligned <- FR_alignment(sound.data, sounds.f0)
#' @section Details:
#' 

FR_alignment <- function(sound.data, sound.f0, cf=1)
{
 
  n <- length(sound.data)
  nc <- dim(sound.data[[1]])[2]
  
  sounds.lengths <- rep(NA, n)
  
  #### time rescaling ###### 
  x <- max(lengths(sound.f0))
  eq.grid <- seq(0, 1, len=x)
  Data.grid <- array(NA, c(n, x, nc))
  f0.grid <- matrix(NA, n, x)
  for (k in 1:n)
  {
    for (i in 1:nc)
    { 
      Data.grid[k, , i] <- approx(seq(0, 1, len=dim(sound.data[[k]])[1]), sound.data[[k]][, i], eq.grid)$y
    }
    f0.grid[k, ] <- approx(seq(0, 1, len=dim(sound.data[[k]])[1]), sound.f0[[k]], eq.grid)$y
    sounds.lengths[k] <- length(sound.f0[[k]])
  }
  
  abs <- seq(0, 1, len=x)
  
  FV <- align_fPCA(t(Data.grid[, , cf]), abs)
  
  aligned.sounds <- array(NA, dim(Data.grid))
  aligned.f0 <- matrix(NA, n, x)
  for (i in 1:n)
  {
    for (j in 1:nc)
    {
      aligned.sounds[i, , j] <- approx(abs, Data.grid[i, , j], xout=FV$gam[, i], rule=2)$y
    }
  }
  
  for (i in 1:n)
  {
    aligned.f0[i, ] <- approx(abs, f0.grid[i, ], xout=FV$gam[, i], rule=2)$y
  }
  
  warping.functions <- t(FV$gam)
  out <- list(mfcc=aligned.sounds, f0=aligned.f0, sounds.lengths=sounds.lengths, warping.functions=warping.functions)
  return(invisible(out))
}


dist_matrix_warp <- function(MFCC)
{
  n <- dim(MFCC$mfcc)[1]
  distance.matrix <- matrix(0, n, n)
  for (k in 2:n)
  {
    for (j in 1:(k-1))
    {
      distance.matrix[k, j] <- dist_mfcc(MFCC$mfcc[k, ,], MFCC$mfcc[j, ,])+acos(cor(MFCC$warping.functions[k, ], MFCC$warping.functions[j, ]))+acos(min(MFCC$sounds.lengths[k], MFCC$sounds.lengths[j])/max(MFCC$sounds.lengths[k], MFCC$sounds.lengths[j]))
      distance.matrix[j, k] <- distance.matrix[k, j]
    }
  }
  return(invisible(distance.matrix))
}


#### MFCCs PCA
#' @examples
#' 
#' @section Details:
#'

## optimize this
mfcc.pca <- function(mfcc, k=1, scaled=TRUE)
{
  n <- dim(mfcc)[1]
  t <- dim(mfcc)[2]
  nc <- dim(mfcc)[3]
  X <- matrix(NA, n, t*nc)
  for (i in 1:n)
  {
    X[i, ] <- as.vector(mfcc[i, ,])
  }
  PCA <- prcomp(X, scale.=scaled)
  PCA_dir <- array(NA, c(k, t, nc))
  for (i in 1:k)
  {
    PCA_dir[i, ,] <- matrix(PCA$rotation[, i], t, nc)
  }
  scaled_scores <- PCA$x[, 1:k]
  scores <- matrix(NA, n, k)
  X <- X-colMeans(X)
  for (i in 1:n)
  {
    for (j in 1:k)
    { 
    scores[i, j] <- t(X[i, ])%*%PCA$rotation[, j]
    }
  }
  out <- list(scores=scores, scaled_scores=scaled_scores, pc=PCA_dir)
  return(invisible(out))
}


## Compute first k PCs of mfccs
mfccs_pc <- function(mfccs, k=1){
    dims <- dim(mfccs)
    mfccs2 <- matrix(mfccs, nrow=dims[1])
    X <- scale(mfccs2, center=TRUE, scale=FALSE)
    svd.loc <- svd::propack.svd(X, neig=k)
    return( list(scores.norm = svd.loc$u * sqrt(dims[1]-1),
                 vars = svd.loc$d^2/(dims[1] - 1),
                 prcomps = array( t(svd.loc$v) , c(k, dims[-1]) )) ) 
}

#### MFCCs LDA
#' @examples
#' 
#' @section Details:
#'

mfcc.lda <- function(mfcc, grouping)
{
  n <- dim(mfcc)[1]
  t <- dim(mfcc)[2]
  nc <- dim(mfcc)[3]
  g <- length(levels(as.factor(grouping)))-1
  X <- matrix(NA, n, t*nc)
  for (i in 1:n){
    X[i, ] <- as.vector(mfcc[i, ,])
  }
  LDA <- lda(X, grouping = grouping)
  LDA_dir <- array(NA, c(g, t, nc))
  for (i in 1:dim(LDA_dir)[1]){ 
  LDA_dir[i, ,] <- matrix(LDA$scaling[, i], t, nc)
  }
  scores <- predict(LDA, dimen=g)$x
  out <- list(scores=scores, ld=LDA_dir)
  return(invisible(out))
}

#### MFCCs LDA of fPCA coeff
#' @examples
#' 
#' @section Details:
#'

mfcc.flda <- function(mfcc, grouping, fk=3)
{
  n <- dim(mfcc)[1]
  t <- dim(mfcc)[2]
  nc <- dim(mfcc)[3]
  g <- length(levels(as.factor(grouping)))-1
  X <- matrix(NA, n, fk*nc)
  V <- matrix(NA, t, fk*nc)
  for (i in 1:nc){
    XF <- mfcc[, , i]
    PC <- prcomp(XF, retx = TRUE)
    X[, (1+(i-1)*fk):(i*fk)] <- PC$x[, 1:fk]
    V[, (1+(i-1)*fk):(i*fk)] <- PC$rotation[, 1:fk]
  }
  
  LDA <- lda(X, grouping = grouping)
  
  scores <- predict(LDA, dimen=g)$x
  out <- list(scores=scores, ld_weights=LDA$scale)
  return(invisible(out))
}



#### map MFCCs coefficients back to the original time scale
#' @examples
#' 
#' @section Details:
#' 

inverse_warp <- function(mfcc, f0=NULL, warping.function, original.length)
{
  warp <- matrix(NA, original.length, dim(mfcc)[2])
  for (j in 1:dim(mfcc)[2]){
    warp[, j] <- approx(warping.function, mfcc[, j], xout=seq(0, 1, len=original.length), rule=2)$y
  }
  
  if (is.null(f0)==FALSE){ 
  warp.f0 <- approx(warping.function, f0, xout=seq(0, 1, len=original.length), rule=2)$y
  sound <- list(mfcc=warp, f0=warp.f0)
  }else{
    sound <- list(mfcc=warp)
  }
  return(invisible(sound))
}

## could be optimized for speed by giving argument at which inverse is to be computed
get_inverse_warp <- function(warping.function)
{
    len <- length(warping.function)
    # approx(seq(0, 1, len=len), warping.function, xout=seq(0, 1, len=original.length), rule=2)$y
    x.out <- seq(0, 1, len=len)
   sapply(x.out, function (y) uniroot((function (x) approx(seq(0, 1, len=len), warping.function, xout=x)$y  - y), lower = 0, upper = 1)[[1]], simplify='array')
}
get_inverse_warp2 <- function(warping.function, x.out)
{
    len <- length(warping.function)
    # approx(seq(0, 1, len=len), warping.function, xout=seq(0, 1, len=original.length), rule=2)$y
   sapply(x.out, function (y) uniroot((function (x) approx(seq(0, 1, len=len), warping.function, xout=x)$y  - y), lower = 0, upper = 1)[[1]], simplify='array')
}


#### average and time realignment of a set of MFCCs
#' @examples
#' 
#' @section Details:
#'

mfcc.mean <- function(aligned.mfcc, aligned.f0, warping.function, length)
{
  n <- dim(aligned.mfcc)[1]
  t <- dim(aligned.mfcc)[2]
  nc <- dim(aligned.mfcc)[3]
  
  mfcc.mean <- apply(aligned.mfcc, c(2, 3), mean)
  unwarp.mean <- inverse_warp(mfcc.mean, f0=aligned.f0, warping.function=warping.function, original.length = length)
  return(invisible(unwarp.mean))
}


#### map the warping function into a linear space
#' @examples
#' 
#' @section Details:
#' t vector of points between 0 and 1 where the warping function is evaluated
#' warping warping function evaluated at the pints in t

warp_map <- function(t, warping)
{
  f <- splinefun(t, warping, method="monoH.FC")
  h <- log(f(t, deriv=1))
  fh <- splinefun(t, h)
  #z <- integrate(fh, 0, 1)$value
  w <- fh(t)#/z
  w
} 

#### map the transformed warping function back to the original space
#' @examples
#' 
#' @section Details:
#'

warp_back <- function(t, w)
{
  f <- splinefun(t, exp(w), method="natural")
  z <- integrate(f, 0, 1)$value
  h <- rep(NA, length(t))
  for (k in 1:length(t)){
    h[k] <- integrate(f, 0, t[k])$value
  }
  h <- h/z
  h
} 

#### Warping functions PCA
#' @examples
#' 
#' @section Details:
#'

warp.pca <- function(warping.functions, k=1, scaled=FALSE)
{
  n <- dim(warping.functions)[1]
  t <- dim(warping.functions)[2]
  time <- seq(0, 1, len=t)
  X <- matrix(NA, n, t)
  for (i in 1:n){
    X[i, ] <- warp_map(time, warping.functions[i, ])
  }
  PCA <- prcomp(X, scale.=scaled)
  PCA_dir <- PCA$rotation[, 1:k]
  scaled_scores <- PCA$x[, 1:k]
  scores <- matrix(NA, n, k)
  X <- X-colMeans(X)
  for (i in 1:n){
    for (j in 1:k){ 
      scores[i, j] <- t(X[i, ])%*%PCA$rotation[, j]
    }
  }
  out <- list(scores=scores, scaled_scores=scaled_scores, pc=PCA_dir)
  return(invisible(out))
}




#### distance matrix between cluster means
#' @examples
#' 
#' @section Details:
#'

means.dist.matrix <- function(aligned.mfcc, warping.function, word.lengths, groups, weights=c(1/3, 1/3, 1/3))
{
  n <- dim(aligned.mfcc)[1]
  t <- dim(aligned.mfcc)[2]
  grid <- seq(0, 1, len=t)
  nc <- dim(aligned.mfcc)[3]
  groups <- factor(groups, levels=unique(groups))
  lv <- levels(groups)
  g <- length(lv)
  means.array <- array(NA, c(g, t, nc))
  warping.matrix <- matrix(NA, g, t)
  lengths.vector <- rep(NA, g)
  for (k in 1:g){
    index <- which(groups==lv[k])
    means.array[k, ,] <- apply(aligned.mfcc[index, ,], c(2, 3), mean)
    warping.matrix[k, ] <- colMeans(warping.function[index, ])
    lengths.vector[k] <- mean(word.lengths[index])
  }
 
  distance.matrix <- matrix(0, g, g)
  for (i in 2:g){
    for (j in 1:(k-1)){
      distance.matrix[i, j] <- weights[1]*dist_mfcc(means.array[i, ,], means.array[j, ,])+weights[2]*acos(abs(cor(warp_map(grid, warping.matrix[i, ]), warp_map(grid, warping.matrix[j, ]))))+weights[3]*acos(min(lengths.vector[i], lengths.vector[j])/max(lengths.vector[i], lengths.vector[j]))
      distance.matrix[j, i] <- distance.matrix[i, j]
    }
  }
  return(invisible(distance.matrix))
}



#### covariance matrix between MFCCs at all time point
#' @examples
#' 
#' @section Details:
#'
cov_process <- function(mfcc, center=TRUE)
{
  n <- dim(mfcc)[1]
  t <- dim(mfcc)[2]
  nc <- dim(mfcc)[3]
  cov.series <- array(NA, c(t, nc, nc))
  if (center==TRUE){ 
  for (k in 1:t){
    cov.series[k, ,] <- cov(mfcc[, k, ])
  }
  }else{
    for (k in 1:t){
    cov.series[k, ,] <- (1/(n-1))*t(mfcc[, k, ])%*%mfcc[, k, ]
    }
  }
  return(invisible(cov.series))
}


#### square root dispersion matrix between MFCCs at all time point
#' @examples
#' 
#' @section Details:
#'
cov_process_sqrt <- function(mfcc, centered=FALSE)
{
    n <- dim(mfcc)[1]
    t <- dim(mfcc)[2]
    nc <- dim(mfcc)[3]
    cov.series <- array(NA, c(t, nc, nc))
    for (k in 1:t){
        cov.series[k, ,] <- cov_sqrt(mfcc[, k, ],centered=centered)
    }
    return(invisible(cov.series))
}



#### distance between covariance time series
#' @examples
#' 
#' @section Details:
#'
## TODO : do we really want to renormalize so that the integral has trace 1, or so that at each 't' the cov has trace 1?
dist_cov_mfcc <- function(cov.series1, cov.series2, dist=sqrtdist,std=FALSE)
{
    t <- dim(cov.series1)[1]
    nc <- dim(cov.series1)[2]
    dt <- rep(NA, t)
    if(std==FALSE){ 
        for (k in 1:t){
            dt[k] <- dist(cov.series1[k, ,], cov.series2[k, ,])^2
        }
    }else{
        trace1 <- apply(cov.series1, 1, function(x) sum(diag(x)))
        cov.series1 <- cov.series1 %>% apply(2:3, function(y) y/trace1)
        trace2 <- apply(cov.series2, 1, function(x) sum(diag(x)))
        cov.series2 <- cov.series2 %>% apply(2:3, function(y) y/trace2)
        # tot.var1<-sum(diag(apply(cov.series1,c(2,3),mean)))
        # tot.var2<-sum(diag(apply(cov.series2,c(2,3),mean)))
        # cov.series1<-cov.series1/tot.var1
        # cov.series2<-cov.series2/tot.var2
        for (k in 1:t){
            dt[k] <- dist(cov.series1[k, ,], cov.series2[k, ,])^2
        }
    }
    return(invisible(sqrt(sum(dt)/t)))
}


#### distance matrix between cluster covariances
#' @examples
#' 
#' @section Details:
#'

cov.dist.matrix <- function(aligned.mfcc, groups, distance=Procdist)
{
  n <- dim(aligned.mfcc)[1]
  t <- dim(aligned.mfcc)[2]
  nc <- dim(aligned.mfcc)[3]
  groups <- factor(groups, levels=unique(groups))
  lv <- levels(groups)
  g <- length(lv)
  cov.array <- array(NA, c(g, t, nc, nc))
 
  for (k in 1:g){
    index <- which(groups==lv[k])
   cov.array[k, ,,] <- cov_process(aligned.mfcc[index, ,])
  }
  
  distance.matrix <- matrix(0, g, g)
  for (i in 2:g){
    for (j in 1:(k-1)){
      distance.matrix[i, j] <- dist_cov_mfcc(cov.array[i, ,,], cov.array[j, ,,], dist=distance)
      distance.matrix[j, i] <- distance.matrix[i, j]
    }
  }
  return(invisible(distance.matrix))
}

#### distance matrix between clusters, taking into account means, covariances and time warpings
#' @examples
#' 
#' @section Details:
#'

phon.dist.matrix <- function(aligned.mfcc, warping.function, word.lengths, groups, weights=c(0.25, 0.25, 0.25, 0.25), distance=Procdist)
{
  D1 <- cov.dist.matrix(aligned.mfcc, groups, distance)
  D2 <- means.dist.matrix(aligned.mfcc, warping.function, word.lengths, groups, weights=weights[1:3]/sum(weights[1:3]))
  distance.matrix <- weights[4]*(D1/max(D1))+sum(weights[1:3])*(D2/max(D2))
  return(invisible(distance.matrix))
}

#' #### pairwise warping and average [this is still work in progress!]
#' #' @examples
#' #' average.mfcc <- pair.average(sound.data[[63:64]], sound.f0[[63:64]], cf=1)
#' #' @section Details:
#' #' 
#' 
#' pair.average <- function(mfcc, f0, dir=getwd()){
#'   aligned.sounds <- FR_alignment(mfcc, f0, cf)
#'   mfcc_avg <- (aligned.sounds$mfcc[1, ,]+aligned.sounds$mfcc[1, ,])/2
#'   mfcc_avg1 <- inverse_warp(mfcc_avg, aligned.sounds$warping.functions[1, ], aligned.sounds$original.length[1])
#'   mfcc_avg2 <- inverse_warp(mfcc_avg, aligned.sounds$f0[1, ], aligned.sounds$warping.functions[2, ], aligned.sounds$original.length[2])
#'   
#'   file <- paste(dir, "/average_1.wav")
#'   mfcc <- list(mfcc=mfcc_avg1$mfcc, f0=f0[[1]])
#'   mfcc_to_sound(mfcc, file)
#'   
#'   file <- paste(dir, "/average_2.wav")
#'   mfcc <- list(mfcc=mfcc_avg2$mfcc, f0=mfcc_avg2$f0)
#'   mfcc_to_sound(mfcc, file)
#'   
#'   file <- paste(dir, "/average_2.wav")
#'   mfcc <- list(mfcc=mfcc_avg, f0=aligned.sounds$f0[1, ])
#'   mfcc_to_sound(mfcc, file)
#'   
#'   return(invisible(mfcc_avg))
#' }
#'   
#'   

#### Frobeniuos distance between two matrices
#' @examples
#' 
#' @section Details:
#'
#'

Fdist <- function(A, B)
{
  sa <- sum((A-B)^2)
  s <- sqrt(sa)
  s
}

#### square root of a positive definite matrix
#' @examples
#' 
#' @section Details:
#'
#'

sqrtM <- function(A){
  
  auto <- eigen(A)
  val <- pmax(auto$values, rep(0, dim(A)[1]))
  d <- sqrt(val)
  D <- diag(d)
  V <- auto$vectors
  L <- V%*%D%*%t(V)
  
  L
}


#### square root dispersion
#' @examples
#' 
#' @section Details:
#'
#'
cov_sqrt <- function(X, centered=FALSE)
{
  N <- dim(X)[1]
  p <- dim(X)[2]
  if (centered==FALSE){ 
    xhat <- colMeans(X)
    X <- sweep(X, 2, xhat)
  }
  
  G <- array(0, c(N, p, p))
  for ( k in 1:dim(G)[1]){
    #G[k,,] <- sqrtM( (X[k,])%*%t(X[k,]) )
    if(sum(X[k,]!=0)){ 
      G[k,,]<-((X[k,])%*%t(X[k,]))/sqrt(as.double(t(X[k,])%*%(X[k,])))
    }
  }
  CG <- apply(G, 2:3, mean)
  C <- CG%*%CG
  return(invisible(C))
}


#### square root dispersion
#' @examples
#' 
#' @section Details:
#'
#'
cov_sqrt_old <- function(X, centered=FALSE)
{
  N <- dim(X)[1]
  p <- dim(X)[2]
  if (centered==FALSE){ 
    xhat <- colMeans(X)
    X <- sweep(X, 2, xhat)
  }
  
  G <- array(0, c(N, p, p))
  for ( k in 1:dim(G)[1]){
    G[k,,] <- sqrtM( (X[k,])%*%t(X[k,]) )
  }
  CG <- apply(G, 2:3, mean)
  C <- CG%*%CG
  return(invisible(C))
}


#### Power matrix of  positive definite matrix
#' @examples
#' 
#' @section Details:
#'
#'

powerM <- function(A, alpha){
    auto <- eigen(A)
    val <- pmax(auto$values, rep(0, dim(A)[1]))
    d <- (val)^(alpha)
    D <- diag(d)
    V <- auto$vectors
    L <- V%*%D%*%t(V)
    L
}

#### Square root distance between two matrices
#' @examples
#' 
#' @section Details:
#'
#'

sqrtdist <- function(A, B){
  sA <- sqrtM(A)
  sB <- sqrtM(B)
  s <- Fdist(sA, sB)
  s
}


#### Procrustes - square root distance between two matrices
#' @examples
#' 
#' @section Details:
#'
#'
#'

Procdist <- function(A, B){
  
  L1 <- sqrtM(A)
  L2 <- sqrtM(B)
  C <- t(L2)%*%L1
  SVD <- svd(C)
  R <- SVD$u%*%t(SVD$v)
  Obj <- L1-L2%*%R
  out <- sqrt(sum( rowSums( t(Obj) * t(Obj))))
  out
}

#### Logarithm of a positive definite matrix
#' @examples
#' 
#' @section Details:
#'
#'

LogM <- function(A){
  
  auto <- eigen(A)
  
  d <- log(auto$values)
  D <- diag(d)
  V <- auto$vectors
  L <- V%*%D%*%solve(V)
  
  L
}

#### Exponential  of a positive definite matrix
#' @examples
#' 
#' @section Details:
#'
#'

ExpM <- function(A){
  
  auto <- eigen(A)
  
  d <- exp(auto$values)
  D <- diag(d)
  V <- auto$vectors
  L <- V%*%D%*%solve(V)
  
  return(L)
}



#### Log-matrix distance between two matrices
#' @examples
#' 
#' @section Details:
#'
#'
#'

LogDist <- function(A, B){
  
  d <- Fnorm(LogM(A)-LogM(B))
  
  return(d)
}



#' do two segments intersect?
#'
#' @examples
#' A <- c(0, 0)
#' B <- c(1, 1)
#' C <- c(.1, .10003)
#' D <- c(.8, 1) 
#' plot( rbind(A, B), type='l')
#' lines(rbind(C, D))
#' seg1 <- rbind(A, B)
#' seg2 <- rbind(C, D)
#' segments_intersect(seg1, seg2)
segments_intersect <- function(seg1, seg2)
{
    A <- cbind(seg2[2, ] - seg2[1, ], seg1[1, ] - seg1[2, ])
    b <- matrix(seg2[2, ] - seg1[2, ], nrow=2)
    tmp <- lm.fit(A, b)$coef
    tmp[is.na(tmp)] <- 0
    all((round(tmp, 15) <= 1) & (round(tmp, 15) >=0))
}



#' Given positions of data observations and a distance matrix between each
#' pairs of observations, output the linear model objects used to derive a
#' density that yields those distances 
distance2densitylm <- function(data.positions, D, border.nodes, extra.nodes=NULL, max.area=NA, min.angle=NA)
{
    data.positions <- as.matrix(data.positions)

    border.segments <- outer(0:nrow(border.nodes), 0:1, '+') %% nrow(border.nodes)  + 1
    #nodes <- rbind(uk.nodes, as.matrix(locations[, 2:3]))
    which.inside <- point.in.polygon(data.positions[, 1], data.positions[, 2], border.nodes[, 1], border.nodes[, 2])
    which.inside <- which(which.inside == 1)
    mesh <- create.MESH.2D(nodes=rbind(border.nodes, as.matrix(data.positions[which.inside, 1:2])), segments=border.segments)
    print(nrow(mesh$nodes))
    mesh <- refine.MESH.2D(mesh, maximum_area = max.area, minimum_angle=min.angle)
    print(nrow(mesh$nodes))
    plot(mesh, xaxt='l', pch=20, cex=.2, yaxt='l')
    points(as.matrix(data.positions[which.inside, 1:2]), pch=20, cex=1)
    #points(data.positions[which.inside, 2:3], col= 1, pch=20)

    D <- D[which.inside, which.inside]
    data.positions <- data.positions[which.inside, ]

    #loc.i <- 1:nrow(D); #1:50; #seq(1, by=2, len=20)
    #D <- D[loc.i, loc.i]
    #data.positions <- data.positions[loc.i, ]

    #points(d.locations[, 2:3], col= 3, pch=as.character(1:18), cex=2)
    fem.basis <- create.FEM.basis(mesh)


    print(paste("Integration of basis along segments will take about ", round(dim(D)[1]^2/2/60*system.time({
        segment = cbind( seq(data.positions[1, 1], data.positions[2, 1], len=seg.res), seq(data.positions[1, 2], data.positions[2, 2], len=seg.res) )
        eval.basis <- R_eval.FEM.basis(fem.basis, segment)
    })[1], 2), " minutes"))



    ptm <- proc.time()
    nnodes <- dim(mesh$nodes)[1]
    non.border.edges <- which(mesh$edgesmarkers[, 1] == 0)
    F <- array(0, c(nnodes, dim(D)))
    ## Evaluate basis functions on segments
    seg.res=100
    # for each pair of points
    for( i in 1:(dim(D)[1]-1) ){
        print(paste("### i = ", i, " / ", dim(D)[1]))
        pb <- txtProgressBar(min = i, max = dim(D)[1], style = 3)
        #    system.time({
        for(j in (i+1):dim(D)[1]){
            setTxtProgressBar(pb, j)
            #            print(paste("j = ", j, " / ", dim(D)[1]))
            segment = cbind( seq(data.positions[i, 1], data.positions[j, 1], len=seg.res), seq(data.positions[i, 2], data.positions[j, 2], len=seg.res) )
            eval.basis <- R_eval.FEM.basis(fem.basis, segment)
            F[, i, j] <- colMeans(eval.basis)*dist(data.positions[c(i, j), ])
            #
        }
        close(pb)
        #    })
    }
    Y = D[upper.tri(D)]
    X = apply(F, 1, function(x) x[upper.tri(x)] )
    X <- Matrix(X, sparse=TRUE)
    stiff <- R_stiff(fem.basis)
    stiff <- Matrix(stiff, sparse=TRUE)
    X.interior <- X[, - which(mesh$nodesmarkers == 1)]
    stiff.interior <- stiff[- which(mesh$nodesmarkers == 1), - which(mesh$nodesmarkers == 1)]

    tot.time <- proc.time()-ptm
    return( mget(c("Y", "X", "X.interior", "stiff", "stiff.interior", "mesh", "tot.time" )) )
}



#### Vectorization of a symmetric matrix
#' @examples
#' 
#' @section Details:
#'
#'
#'

vec <- function(D){
  d <- rep(0, dim(D)[1]*(dim(D)[1]+1)/2)
  k <- 1
  for (i in 1:dim(D)[1]){
    for (j in i:dim(D)[1]){
      d[k] <- D[i, j]
      k <- k+1
    }
  }
  
  d
}

#### Build a symmetric matrix from the vector of its unique elements
#' @examples
#' 
#' @section Details:
#'
#'
#'

ivec <- function(d){
  n <- (sqrt(8*length(d)+1)-1)/2
  D <- matrix(0, n, n)
  k <- 1
  for (i in 1:n){
    for (j in i:n){
      D[i, j] <- d[k]
      D[j, i] <- d[k]
      k <- k+1
    }
  }
  D
}



#### Projection of the covariance process into the space of symmetric matrices
#' @examples
#' 
#' @section Details:
#'
#'
#'

sym.process <- function(Ct, t.map=sqrtM)
{
  t <- dim(Ct)[1]
  nc <- dim(Ct)[2]
  sym.series <- matrix(NA, t, (nc*(nc+1))/2)
  for (k in 1:t){
    sym.series[k, ] <- vec(t.map(Ct[k, ,]))
  }
  return(invisible(sym.series))
}


#### Great circle distance between two geographical points
#' @examples
#' 
#' @section Details: c1=(Latitude, Longitude) of the first point
#'                   c2=(Latitude, Longitude) of the second point
#'
#'

Geodist <- function(c1, c2)
{
  
  R <- 6352
  r <- (pi/2)/90
  
  d <- as.double(R*acos(sin(c1[1]*r)*sin(c2[1]*r)+cos(c1[1]*r)*cos(c2[1]*r)*cos(-c2[2]*r+c1[2]*r)))
  return(d) 
}



## Plot FEM object as 2D image
FEM.image = function(FEM, res=50, data.locations=NA, border.nodes=NULL, doplot=TRUE, ...)  
{
    # PLOT  Plots a FEM object FDOBJ over a rectangular grid defined by 
    # vectors X and Y;
    #

    #   if (!is.fd(fdobj))
    #   {
    #     stop('FDOBJ is not an FD object')
    #   }
    mesh <- FEM$FEMbasis$mesh

    ranges <- apply(FEM$FEMbasis$mesh$nodes, 2, range)
    x <- seq(from=ranges[1, 1], to=ranges[2, 1], len=res)
    y <- seq(from=ranges[1, 2], to=ranges[2, 2], len=res)

    grid.tmp <- (expand.grid(x, y))

    if(!is.null(border.nodes)){
    which.inside <- point.in.polygon(grid.tmp[,1], grid.tmp[, 2], border.nodes[,1], border.nodes[,2])
    grid.tmp <- grid.tmp[which.inside > 0, ]
    }

#image(acast(data.frame(tmp, which.inside), Var1~Var2, value.var='which.inside'))

    z <- eval.FEM(FEM, as.matrix(grid.tmp)) 
    data.tmp <- data.frame(grid.tmp, z=z)
    library(reshape2)
    z2 <- acast(data.tmp, Var1~Var2, value.var='z')
    x <- as.numeric(dimnames(z2)[[1]])
    y <- as.numeric(dimnames(z2)[[2]])
#   image(x, y, z2)
    #image.plot(x, y, z2, plot.axes=points(data.locations, pch=20), ...)
    if(doplot){
        image.plot(x, y, z2, ...)
        if (!is.null(data.locations) && !is.na(data.locations) ) points(data.locations, pch=20, ...)
    }
    return(invisible(mget(c('x', 'y', 'z2'))))

}

library(fdasrvf)
#### Smoothing and alignment of Mel Frequency Coepstral Coefficients
#' @examples
#' 
#' @section Details:
#'
#' sound.data: a list with the mfccs
#' sound.f0: a list with the f0s

mfcc_preprocess <- function(sound.data, sound.f0, cf=1, resolution=NA, ...)
{
  
  n <- length(sound.data)
  nc <- dim(sound.data[[1]])[2]
  
  sounds.lengths <- rep(NA, n)
  
  #### time rescaling ###### 
  if( is.na(resolution) ){
      x <- max(lengths(sound.f0))
  } else {
      x  <-  resolution
  }
  eq.grid <- seq(0, 1, len=x)
  Data.grid <- array(NA, c(n, x, nc))
  f0.grid <- matrix(NA, n, x)

# create progress bar

  print('Smoothing the functions')
  pb <- txtProgressBar(min = 0, max = n, style = 3)
  for (k in 1:n){
      setTxtProgressBar(pb, k)
      for (i in 1:nc){ 
          #print(paste(k, i, sep='-'))
          fit <- smooth.spline(seq(0, 1, len=dim(sound.data[[k]])[1]), sound.data[[k]][, i], cv=TRUE)
          Data.grid[k, , i] <- predict(fit, eq.grid)$y
      }
      fit <- smooth.spline(seq(0, 1, len=dim(sound.data[[k]])[1]), sound.f0[[k]], cv=TRUE)
      f0.grid[k, ] <- predict(fit, eq.grid)$y
      sounds.lengths[k] <- length(sound.f0[[k]])
  }
  close(pb)
  
  abs <- seq(0, 1, len=x)
  
  FV <- align_fPCA(t(Data.grid[, , cf]), abs, ...)
  
  aligned.sounds <- array(NA, dim(Data.grid))
  aligned.f0 <- matrix(NA, n, x)
  for (i in 1:n){
    for (j in 1:nc){
      aligned.sounds[i, , j] <- approx(abs, Data.grid[i, , j], xout=FV$gam[, i], rule=2)$y
    }
  }
  
  for (i in 1:n){
    aligned.f0[i, ] <- approx(abs, f0.grid[i, ], xout=FV$gam[, i], rule=2)$y
  }
  
  warping.functions <- t(FV$gam)
  out <- list(mfcc=aligned.sounds, f0=aligned.f0, sounds.lengths=sounds.lengths, warping.functions=warping.functions)
  return(invisible(out))
}

## smooth the duration spatially
mean_smooth_duration <- function(prepared.data, h, NN=4, bandwidth.mode, output = 'estimation',
                                 fit.degree, dist.mesh=NULL) #bandwidth.mode='location.NN',
{
    attach(prepared.data)
    ## compute means per locations and the weight scaling factor
    weight.scaling <- numeric(nrow(loc.coords))
    duration.local.means <- array(NA, c(nrow(loc.coords)))
    for(i in 1:nrow(loc.coords))
    {
        ## get which observations correspond to location i
        ind.tmp <- which(obs.coords[,1] == loc.coords[i,1] & obs.coords[,2] == loc.coords[i,2])
        length(ind.tmp) -> weight.scaling[i] ## how many observations for that location
        mean( duration[ind.tmp] ) -> duration.local.means[i]
    }
    rm(duration, obs.coords)

#    browser()
    if(is.null(dist.mesh)){ 
        graph.weights <- sapply(1:nrow(mesh$edges), function(j) dist(mesh$nodes[mesh$edges[j, ],], method='euclidean'))
        g2 <- add.edges(graph.empty(nrow(mesh$nodes)), t(mesh$edges), weight=graph.weights )
        dist.mesh <- shortest.paths(g2, mode='all')
    }
    ##
    #n.obs <- dim(mfcc.data)[1]

    n.loc <- nrow(loc.coords)
    loc.nodes <- rep(NA, n.loc)
    for (i in 1:n.loc){ 
        loc.nodes[i] <-  which(mesh$nodes[, 1]==loc.coords[i, 1]& mesh$nodes[, 2]==loc.coords[i, 2])
    }

    total.variance <- rep(1, n.loc)

    if( output == 'cross-validation' )
    {
        mean.cv <- numeric(nrow(loc.coords))
        for( k.ind in 1:nrow(loc.coords) )
        {
            k <- loc.nodes[k.ind]

            tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode, weight.scaling=weight.scaling[-k.ind], local.dists=dist.mesh[k, loc.nodes[-k.ind]], fit.degree)

            ## compute pointwise fit
            lfit <- compute_pointwise_mean_fit_duration(tau, duration.local.means[-k.ind], current.coord=mesh$nodes[k,], loc.coords[-k.ind,], weight.scaling=weight.scaling[-k.ind], total.variance=total.variance[-k.ind], local.dists=dist.mesh[k, loc.nodes[-k.ind]], fit.degree=fit.degree)

            ## store results
            mean.cv[k.ind] <- sum((lfit$smoothed.duration - duration.local.means[k.ind, ,])^2)/total.variance[k.ind]
        }
        return(mean.cv)
    } else 
    {
        ## set output.indices according to the output
        if( is.matrix(output) )
        {
            output.indices <- numeric(nrow(output))
            for (i in 1:dim(output)[1])
                output.indices[i] <- which(mesh$nodes[, 1]==output[i, 1] & mesh$nodes[, 2]==output[i, 2])
        }  else if (output == 'estimation')
        {
            print(paste0("duration field estimation "))
            output.indices  <- 1:nrow(mesh$nodes)
        } else if(output == 'prediction')
        {
            print(paste0("*duration field prediction "))
            output.indices <- loc.nodes
        }
        else
            stop(sprintf('undefined value or type of \'output\' %s', output))

        n <- length(output.indices)

        duration.smoothed <- array(NA, c(n), dimnames=list(output.indices) )
        effective.number.obs <- numeric(n)
        taus <- numeric(n)
        names(effective.number.obs) <- output.indices


        pb <- txtProgressBar(1, 1+n, style=1)
        for (k.ind in seq(along=output.indices)){
            k <- output.indices[k.ind]

            ## compute bandwidth
            tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode, weight.scaling=weight.scaling, local.dists=dist.mesh[k, loc.nodes], fit.degree)

            ## compute pointwise fit
            lfit <- compute_pointwise_mean_fit_duration(tau,
                                                        duration.local.means,
                                                        current.coord=mesh$nodes[k,],
                                                        loc.coords,
                                                        weight.scaling=weight.scaling,
                                                        total.variance=total.variance,
                                                        local.dists=dist.mesh[k, loc.nodes],
                                                        fit.degree=fit.degree)

            ## store results
            duration.smoothed[k.ind] <- lfit$smoothed.duration
            effective.number.obs[k.ind] <- lfit$effective.number.obs
            taus[k.ind] <- lfit$tau

            setTxtProgressBar(pb, k.ind)
        }
        close(pb)
    on.exit(detach(prepared.data))
        out <- c(list(duration=duration.smoothed, nobs.per.loc=weight.scaling), mesh=mesh, mget(c( 'h', 'NN', 'effective.number.obs', 'taus', 'bandwidth.mode', 'fit.degree'))) 
        return(invisible(out))
    }
}

mean_smooth <- function(mfcc.data, cov.data, obs.coords, loc.coords, output = 'estimation', h, mesh,  dist.mesh=NULL, NN=4, bandwidth.mode, fit.degree) #bandwidth.mode='location.NN', fit.degree=1)
{
    ## compute means per locations and the weight scaling factor
    weight.scaling <- numeric(nrow(loc.coords))
    mfcc.data.local.means <- array(NA, c(nrow(loc.coords), dim(mfcc.data)[-1]))
    for(i in 1:nrow(loc.coords))
    {
        ## get which observations correspond to location i
        ind.tmp <- which(obs.coords[,1] == loc.coords[i,1] & obs.coords[,2] == loc.coords[i,2])
        length(ind.tmp) -> weight.scaling[i] ## how many observations for that location
        apply( mfcc.data[ind.tmp,,], 2:3, mean ) -> mfcc.data.local.means[i,,]
    }
    rm(mfcc.data, obs.coords)
    ##
    #n.obs <- dim(mfcc.data)[1]
    if(is.null(dist.mesh)){ 
        graph.weights <- sapply(1:nrow(mesh$edges), function(j) dist(mesh$nodes[mesh$edges[j, ],], method='euclidean'))
        g2 <- add.edges(graph.empty(nrow(mesh$nodes)), t(mesh$edges), weight=graph.weights )
        dist.mesh <- shortest.paths(g2, mode='all')
    }

    n.loc <- nrow(loc.coords)
    loc.nodes <- rep(NA, n.loc)
    for (i in 1:n.loc){ 
        loc.nodes[i] <-  which(mesh$nodes[, 1]==loc.coords[i, 1]& mesh$nodes[, 2]==loc.coords[i, 2])
    }

    total.variance <- sapply(1:dim(cov.data)[1], function(i) sum(diag(apply(cov.data[i, ,,], c(2, 3), mean))))

    if( output == 'cross-validation' )
    {
        mean.cv <- numeric(nrow(loc.coords))
        for( k.ind in 1:nrow(loc.coords) )
        {
            k <- loc.nodes[k.ind]

            tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode, weight.scaling=weight.scaling[-k.ind], local.dists=dist.mesh[k, loc.nodes[-k.ind]], fit.degree)

            ## compute pointwise fit
            lfit <- compute_pointwise_mean_fit(tau, mfcc.data.local.means[-k.ind,,], current.coord=mesh$nodes[k,], loc.coords[-k.ind,], weight.scaling=weight.scaling[-k.ind], total.variance=total.variance[-k.ind], local.dists=dist.mesh[k, loc.nodes[-k.ind]], fit.degree=fit.degree)

            ## store results
            mean.cv[k.ind] <- sum((lfit$smoothed.mfcc - mfcc.data.local.means[k.ind, ,])^2)/(dim(mfcc.data)[2])/total.variance[k.ind]
        }

        return(mean.cv)

    } else 
    {
        ## set output.indices according to the output
        if( is.matrix(output) )
        {
            output.indices <- numeric(nrow(output))
            for (i in 1:dim(output)[1])
                output.indices[i] <- which(mesh$nodes[, 1]==output[i, 1] & mesh$nodes[, 2]==output[i, 2])
        }  else if (output == 'estimation')
        {
            print(paste0("*mean estimation "))
            output.indices  <- 1:nrow(mesh$nodes)
        } else if(output == 'prediction')
        {
            print(paste0("*mean prediction "))
            output.indices <- loc.nodes
        }
        else
            stop(sprintf('undefined value or type of \'output\' %s', output))

        n <- length(output.indices)

        new.mfcc <- array(NA, c(n, dim(mfcc.data)[2], dim(mfcc.data)[3]), dimnames=list(output.indices) )
        effective.number.obs <- numeric(n)
        taus <- numeric(n)
        names(effective.number.obs) <- output.indices


        pb <- txtProgressBar(1, 1+n, style=1)
        for (k.ind in seq(along=output.indices)){
            k <- output.indices[k.ind]

            ## compute bandwidth
            tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode, weight.scaling=weight.scaling, local.dists=dist.mesh[k, loc.nodes], fit.degree)

            ## compute pointwise fit
            lfit <- compute_pointwise_mean_fit(tau, mfcc.data.local.means, current.coord=mesh$nodes[k,], loc.coords, weight.scaling=weight.scaling, total.variance=total.variance, local.dists=dist.mesh[k, loc.nodes], fit.degree=fit.degree)

            ## store results
            new.mfcc[k.ind, ,] <- lfit$smoothed.mfcc
            effective.number.obs[k.ind] <- lfit$effective.number.obs
            taus[k.ind] <- lfit$tau

            setTxtProgressBar(pb, k.ind)
        }
        close(pb)
        out <- c(list(mfcc=new.mfcc, nobs.per.loc=weight.scaling), mget(c('mesh', 'h', 'NN', 'effective.number.obs', 'taus', 'bandwidth.mode', 'fit.degree'))) 
        return(invisible(out))
    }
}


## prepare mfccs for smoothing
## distance.method='fast' : approximates geodesic distance by graph distance
## distance.method='precise' : approximates geodesic distance by graph distance if segment is not entirely inside the domain, otherwise uses the distance of the segment
## 
prepare_mfccs <- function(mfcc.data, local.bnc, border.nodes, min.angle=NA, max.area=NA, distance.method='fast')
{
    loc.names <- unique(local.bnc$placenamecleaned)
    locations_coord <- matrix(NA, length(loc.names), 2)
    row.names(locations_coord)<- loc.names
    for (k in 1:length(loc.names)){
        index <- which(local.bnc$placenamecleaned==loc.names[k])[1]
        locations_coord[k, ] <- bnc_to_lonlat(local.bnc[index, ])
    }

    ##

    # ###### observations locations #####
    Lon <- rep(0, dim(mfcc.data)[1])
    Lat <- rep(0, dim(mfcc.data)[1])
    for (k in 1:length(Lon)){
        index <- which(loc.names==local.bnc$placenamecleaned[k])
        Lon[k] <- locations_coord[index, 1]
        Lat[k] <- locations_coord[index, 2]
    }

    ### sample covariances and locations ####
    nobs.loc <- rep(NA, dim(locations_coord)[1])
    cov.set <- array(NA, c(dim(locations_coord)[1], dim(mfcc.data)[2], dim(mfcc.data)[3], dim(mfcc.data)[3]))
    rm.index <- c()
    for (k in 1:dim(locations_coord)[1]){
        index <- which(Lon==locations_coord[k, 1] & Lat==locations_coord[k, 2])
        if (length(index)>4){ 
            cov.set[k, ,,] <- cov_process(mfcc.data[index, ,])
            nobs.loc[k] <- length(index)
        }else{
            rm.index <- c(rm.index, k)
        }
    }

    #### locations subset ###
    locations_cov <- locations_coord[-rm.index, ]
    cov.set <- cov.set[-rm.index, ,,]
    nobs.loc <-nobs.loc[-rm.index]
    data.positions <- unique(locations_cov)
    ####

    border.segments <- outer(0:nrow(border.nodes), 0:1, '+') %% nrow(border.nodes)  + 1
    #nodes <- rbind(uk.nodes, as.matrix(locations[, 2:3]))
    which.inside <- point.in.polygon(data.positions[, 1], data.positions[, 2], border.nodes[, 1], border.nodes[, 2])
    which.inside <- which(which.inside == 1)
    mesh <- create.MESH.2D(nodes=rbind(border.nodes, as.matrix(data.positions[which.inside, 1:2])), segments=border.segments)
    mesh <- refine.MESH.2D(mesh, maximum_area = max.area, minimum_angle=min.angle)
    print(sprintf('Mesh with %d nodes', mesh$nodes %>% nrow))

    cov.data <- cov.set[which.inside, ,,]
    loc.coords <- locations_cov[which.inside, ]
    n.sd <- nobs.loc[which.inside]
    rm.data <- which(!(Lon%in%loc.coords[, 1] & Lat%in%loc.coords[, 2]))
    mfcc.data <- mfcc.data[-rm.data, ,]
    obs.coords <- cbind(Lon[-rm.data], Lat[-rm.data])
    duration <- local.bnc[['vowel.duration']][-rm.data]
    local.bnc <- local.bnc[-rm.data,]


    n.obs <- dim(mfcc.data)[1]
    n.loc <- dim(cov.data)[1]
    n <- dim(mesh$nodes)[1]
    new.mfcc <- array(NA, c(n, dim(mfcc.data)[2], dim(mfcc.data)[3]))
    new.cov <- array(NA, c(n, dim(cov.data)[2], dim(cov.data)[3], dim(cov.data)[4]))

    ## computing the pairwise distances
    dist.mesh = switch(distance.method,
                       precise = approx_geodesic_distances_precise(mesh, border.nodes, loc.coords, method='euclidean'), 
                       fast = approx_geodesic_distances_fast(mesh, border.nodes, method='euclidean')
                       )

    return( invisible( mget(c('mfcc.data', 'cov.data', 'obs.coords', 'loc.coords', 'duration', 'mesh', 'dist.mesh', 'n.sd', 'border.nodes', 'local.bnc')) ) )

}

## function to approximate geodesic distances. If the segment between two nodes
## is inside the domain, then the geodesic distance is the segment length,
## otherwise it is approximated by the graph distance
## NOTE: maybe only the distances from the observation locations are needed!
## NEEDS DEBUGGING
approx_geodesic_distances_precise <- function(mesh, border.nodes, loc.coords, method='euclidean'){
    loc.nodes <- rep(NA, nrow(loc.coords))
    for (i in seq(loc.nodes)){ 
        loc.nodes[i] <-  which(mesh$nodes[, 1]==loc.coords[i, 1]& mesh$nodes[, 2]==loc.coords[i, 2])
    }
    #
    graph.weights  <-  sapply(1:nrow(mesh$edges), function(j) dist(mesh$nodes[mesh$edges[j, ],], method='euclidean'))
    g2  <-  add.edges(graph.empty(nrow(mesh$nodes)), t(mesh$edges), weight=graph.weights )
    graph.dist  <-  shortest.paths(g2, mode='all')
    rm(g2)
    dist.mesh <- matrix(NA, nrow(mesh$nodes), nrow(mesh$nodes))
    diag(dist.mesh) <- 0
    border.polygon <- readWKT(sprintf("POLYGON((%s))", paste(apply(rbind(border.nodes, border.nodes[1,]), 1, paste, collapse=' '), collapse=",")))

    for(idx in seq(loc.nodes)){
        i=loc.nodes[idx]
        print(sprintf("%d/%d", idx, length(loc.nodes)))
        for(j in 1:nrow(dist.mesh)){
            current.segment <- readWKT(sprintf("LINESTRING(%s)", paste(apply(mesh$nodes[c(i,j),], 1, paste, collapse=' '), collapse=", ")))
            if( gCovers(border.polygon, current.segment) ){ ## segment is inside the domain
                dist.mesh[i,j] <- dist.mesh[j,i]  <- dist(mesh$nodes[c(i,j),], method=method)
            } else { ## the segment crossed the border
              #  print('graph dist approximation')
                dist.mesh[i,j] <- dist.mesh[j,i] <- graph.dist[i,j]
            }
        }
    }
    return(dist.mesh)
}

## OLD function to approximate the geodesic distances as graph distances
approx_geodesic_distances_fast <- function(mesh, border.nodes, method='euclidean'){
    graph.weights  <-  sapply(1:nrow(mesh$edges), function(j) dist(mesh$nodes[mesh$edges[j, ],], method=method))
    g2  <-  add.edges(graph.empty(nrow(mesh$nodes)), t(mesh$edges), weight=graph.weights )
    dist.mesh  <-  shortest.paths(g2, mode='all')
    return(dist.mesh)
}

#### pyramidal estimation of smooth mean and variance # new version for the revision
#' @examples
#' 
#' @section Details:
#'
#'
#' cov.smooth.type = c('l2', 'sqrt')
## TODO: 
smooth_global_new <- function(prepared.data,  paral=NA,
                           mean.h,
                           mean.nn,
                           mean.bandwidth.mode='location.NN', 
                           niter = 0,
                           fit.degree, 
                           cov.h,
                           cov.nn,
                           cov.smooth.type,
                           cov.bandwidth.mode='location.NN',
                           alpha=NA)
{
    ## set which covariance smoothing we are using
    cov_smooth <- get(paste0('cov_smooth_', cov.smooth.type))

    ## iterate prediction of mean and covariance
    if( niter > 0)
    {
        for( iter in 1:niter )
        {
            mean.field <- prepared.data %$% mean_smooth(mfcc.data, cov.data=cov.data, obs.coords=obs.coords, loc.coords=loc.coords, output='prediction', h=mean.h, mesh=mesh, dist.mesh,  NN=mean.nn, bandwidth.mode=mean.bandwidth.mode, fit.degree=fit.degree) 
            cov.field <- prepared.data %$% cov_smooth(mfcc.data, obs.coords, loc.coords=loc.coords, output='prediction', mfcc.center=mean.field[['mfcc']], h=cov.h, mesh, n.sd=n.sd, dist.mesh, NN=cov.nn, fit.degree=fit.degree, bandwidth.mode=cov.bandwidth.mode, alpha=alpha)
            cov.data <- cov.field[['cov']]
        }
    }
    mean.field <- prepared.data %$% mean_smooth(mfcc.data, cov.data=cov.data, obs.coords, loc.coords=loc.coords, output='estimation', h=mean.h, mesh=mesh, dist.mesh,  NN=mean.nn, bandwidth.mode=mean.bandwidth.mode, fit.degree=fit.degree) 
    duration.field <- mean_smooth_duration(prepared.data, h=mean.h, NN=mean.nn, bandwidth.mode=mean.bandwidth.mode, fit.degree=fit.degree) 

    warning('computing cov field without using mean field estimate')
    #cov.field <- list(cov=NA, taus=NA, effective.number.obs=NA)
    cov.field <- prepared.data %$% cov_smooth(mfcc.data, obs.coords=obs.coords, loc.coords=loc.coords, output='estimation', mfcc.center=NULL, h=cov.h, mesh, n.sd=n.sd, dist.mesh, NN=cov.nn, fit.degree=fit.degree, bandwidth.mode=cov.bandwidth.mode, alpha=alpha)


    pca <- mfccs_pc(prepared.data$mfcc.data, k=10)

    out <- c(list(mfcc=mean.field$mfcc, cov=cov.field$cov,
                  duration=duration.field$duration, mean.taus=mean.field$tau,
                  cov.taus=cov.field$taus, nobs.per.loc=mean.field$nobs.per.loc,
                  mean.effective.nobs=mean.field$effective.number.obs,
                  cov.effective.nobs=cov.field$effective.number.obs),  
             prepared.data,
             mget(c( paste0('mean.', c('h', 'nn', 'bandwidth.mode')),
                    paste0('cov.', c('h', 'nn', 'bandwidth.mode')),
                    'fit.degree',  'pca', 'cov.smooth.type'))) 
    #mget(c('mesh', 'h', 'NN', 'bandwidth.mode', 'fit.degree', 'obs.coords',
    #'loc.coords', 'mfcc.data', 'pca', 'border.nodes')))
    class(out) <- 'smooth_field'
    return(invisible(out))
}

#### OLD: pyramidal estimation of smooth mean and variance
#' @examples
#' 
#' @section Details:
#'
#'
#' cov.smooth.type = c('l2', 'sqrt')
smooth_global <- function(mfcc.data, local.bnc, border.nodes,  paral=NA,
                           mean.h,
                           mean.nn,
                           mean.bandwidth.mode='location.NN', 
                           niter = 0,
                           fit.degree, min.angle=NA, max.area=NA,
                           cov.h,
                           cov.nn,
                           cov.smooth.type,
                           cov.bandwidth.mode='location.NN',
                           alpha=NA)
{
    prepared.data <- prepare_mfccs(mfcc.data, local.bnc, border.nodes, min.angle, max.area)
    ## set which covariance smoothing we are using
    cov_smooth <- get(paste0('cov_smooth_', cov.smooth.type))

    ## iterate prediction of mean and covariance
    if( niter > 0)
    {
        for( iter in 1:niter )
        {
            mean.field <- prepared.data %$% mean_smooth(mfcc.data, cov.data=cov.data, obs.coords=obs.coords, loc.coords=loc.coords, output='prediction', h=mean.h, mesh=mesh, dist.mesh,  NN=mean.nn, bandwidth.mode=mean.bandwidth.mode, fit.degree=fit.degree) 
            cov.field <- prepared.data %$% cov_smooth(mfcc.data, obs.coords, loc.coords=loc.coords, output='prediction', mfcc.center=mean.field[['mfcc']], h=cov.h, mesh, n.sd=n.sd, dist.mesh, NN=cov.nn, fit.degree=fit.degree, bandwidth.mode=cov.bandwidth.mode, alpha=alpha)
            cov.data <- cov.field[['cov']]
        }
    }
    mean.field <- prepared.data %$% mean_smooth(mfcc.data, cov.data=cov.data, obs.coords, loc.coords=loc.coords, output='estimation', h=mean.h, mesh=mesh, dist.mesh,  NN=mean.nn, bandwidth.mode=mean.bandwidth.mode, fit.degree=fit.degree) 
    warning('computing cov field without using mean field estimate')
    #cov.field <- list(cov=NA, taus=NA, effective.number.obs=NA)
    cov.field <- prepared.data %$% cov_smooth(mfcc.data, obs.coords=obs.coords, loc.coords=loc.coords, output='estimation', mfcc.center=NULL, h=cov.h, mesh, n.sd=n.sd, dist.mesh, NN=cov.nn, fit.degree=fit.degree, bandwidth.mode=cov.bandwidth.mode, alpha=alpha)
    #cov.field <- prepared.data %$% cov_smooth(mfcc.data, obs.coords=obs.coords, loc.coords=loc.coords, output='estimation', mfcc.center=mean.field$mfcc, h[2], mesh, n.sd, dist.mesh, NN=NN, fit.degree=fit.degree, bandwidth.mode=bandwidth.mode)



    pca <- mfccs_pc(prepared.data$mfcc.data, k=10)

    out <- c(list(mfcc=mean.field$mfcc, cov=cov.field$cov, mean.taus=mean.field$tau,
                  cov.taus=cov.field$taus, nobs.per.loc=mean.field$nobs.per.loc,
                  mean.effective.nobs=mean.field$effective.number.obs,
                  cov.effective.nobs=cov.field$effective.number.obs),  
             prepared.data,
             mget(c( paste0('mean.', c('h', 'nn', 'bandwidth.mode')),
                    paste0('cov.', c('h', 'nn', 'bandwidth.mode')),
                    'fit.degree',  'pca', 'border.nodes', 'cov.smooth.type'))) 
    #mget(c('mesh', 'h', 'NN', 'bandwidth.mode', 'fit.degree', 'obs.coords',
    #'loc.coords', 'mfcc.data', 'pca', 'border.nodes')))
    class(out) <- 'smooth_field'
    return(invisible(out))
}

## smooth the covariance field using the Euclidean metric
cov_smooth_l2 <-function(mfcc.data, obs.coords, loc.coords, mfcc.center=NULL, h,mesh, n.sd=rep(1, dim(loc.coords)[1]), dist.mesh=NULL, NN=4, output = 'estimation', bandwidth.mode, fit.degree, alpha=NA) #bandwidth.mode='location.NN', fit.degree=1)
{
    n.obs <- dim(mfcc.data)[1]
    n.loc <- dim(loc.coords)[1]

    if(is.null(dist.mesh)){ 
        graph.weights <- sapply(1:nrow(mesh$edges), function(j) dist(mesh$nodes[mesh$edges[j, ],], method='euclidean'))
        g2 <- add.edges(graph.empty(n), t(mesh$edges), weight=graph.weights )
        dist.mesh <- shortest.paths(g2, mode='all')
    }
    obs.nodes <- rep(NA, n.obs)
    for (i in 1:n.obs){ 
        obs.nodes[i] <- which(mesh$nodes[, 1]==obs.coords[i, 1]& mesh$nodes[, 2]==obs.coords[i, 2])
    }

    loc.nodes <- rep(NA, n.loc)
    for (i in 1:n.loc){ 
        loc.nodes[i] <- which(mesh$nodes[, 1]==loc.coords[i, 1]& mesh$nodes[, 2]==loc.coords[i, 2])
    }

    ctr.cov <- array(NA, c(n.loc, dim(mfcc.data)[2], dim(mfcc.data)[3], dim(mfcc.data)[3]))
    if (is.null(mfcc.center)){
        for (k in 1:n.loc){
            index <- which(mesh$nodes[obs.nodes, 1]==loc.coords[k, 1]& mesh$nodes[obs.nodes, 2]==loc.coords[k, 2])
            ctr.cov[k, ,,] <- cov_process(mfcc.data[index, ,], center=TRUE)
        }
    }else{
        mfcc2 <- array(NA, c(n.obs, dim(mfcc.data)[2], dim(mfcc.data)[3]))
        for (i in 1:n.obs){
            index <- which(loc.coords[, 1]==mesh$nodes[obs.nodes[i], 1] & loc.coords[, 2]==mesh$nodes[obs.nodes[i], 2])[1]
            mfcc2[i, ,] <- mfcc.data[i, ,]-mfcc.center[index, ,]
        }
        for (k in 1:n.loc){
            index <- which(obs.nodes==loc.nodes[k])
            ctr.cov[k, ,,] <- cov_process(mfcc2[index, ,], center=FALSE)
        }
    }


    ## set output.indices according to the output
    if( output == 'cross-validation' )
    {
        ##
        cov.cv <- numeric(nrow(loc.coords))
        for( k.ind in 1:nrow(loc.coords) )
        {
            k <- loc.nodes[k.ind]

            ## for L2 smoothing, we use the same bandwidth function as for the mean
            tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode, weight.scaling= n.sd[-k.ind], local.dists=dist.mesh[k, loc.nodes[-k.ind]], fit.degree)

            ## compute pointwise fit
            lfit <- compute_pointwise_cov_fit(tau, ctr.cov[-k.ind, ,,],
                                              current.coord=mesh$nodes[k,],
                                              loc.coords=loc.coords[-k.ind,],
                                              weight.scaling=n.sd[-k.ind],
                                              local.dists=dist.mesh[k, loc.nodes[-k.ind]], fit.degree=fit.degree)

            cov.predict <- lfit$smoothed.logcov

            ## store results
            cov.cv[k.ind] <- apply(cov.predict - ctr.cov[k.ind,,,], 1, Matrix::norm, 'F')^2 %>% mean
            #dist_cov_mfcc(cov.predict, ctr.cov[k.ind,,,], dist=sqrtdist)^2
        }

        return(cov.cv)

    } else 
    {
        if( is.matrix(output) )
        {
            output.indices <- numeric(nrow(output))
            for (i in 1:dim(output)[1])
                output.indices[i] <- which(mesh$nodes[, 1]==output[i, 1] & mesh$nodes[, 2]==output[i, 2])
        }  else if (output == 'estimation')
        {
            print(paste0("cov estimation"))
            output.indices  <- 1:nrow(mesh$nodes)
        } else if(output == 'prediction')
        {
            print(paste0("*cov prediction "))
            output.indices <- loc.nodes
        }
        else
            stop(sprintf('undefined value or type of \'output\' %s', output))

        n <- length(output.indices)

        obs.cov <- array(NA, c(n, dim(ctr.cov)[-1]), dimnames=list(output.indices=output.indices))
        taus <- numeric(n)
        effective.number.obs <- numeric(n)

        pb <- txtProgressBar(1, 1+n, style=1)

        for (k.ind in seq(along=output.indices)){
            k <- output.indices[k.ind]

            tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode = bandwidth.mode, weight.scaling = n.sd, local.dists=dist.mesh[k, loc.nodes], fit.degree)


            ## compute pointwise fit
            lfit <- compute_pointwise_cov_fit(tau, ctr.cov,
                                              current.coord=mesh$nodes[k,],
                                              loc.coords=loc.coords,
                                              weight.scaling=n.sd,
                                              local.dists=dist.mesh[k, loc.nodes], fit.degree=fit.degree)


            ## store results
            obs.cov[k.ind, , ,] <- lfit$smoothed.logcov
            taus[k.ind] <- lfit$tau ## tau might have been modified by compute_pointwise_cov_fit
            effective.number.obs[k.ind] <- lfit$effective.number.obs ## tau might have been modified by compute_pointwise_cov_fit

            setTxtProgressBar(pb, k.ind)
        }

        out <- c(list(cov=obs.cov), mget(c('mesh', 'h', 'NN', 'taus', 'bandwidth.mode', 'fit.degree', 'effective.number.obs')))
        return(invisible(out))
    }
}

## Smooth the ds-covariance field using the ds metric
cov_smooth_sqrt <-function(mfcc.data, obs.coords, loc.coords, mfcc.center=NULL, h, mesh, n.sd=rep(1, dim(loc.coords)[1]), dist.mesh=NULL, NN=4, output = 'estimation', bandwidth.mode, fit.degree, alpha=NA) #bandwidth.mode='location.NN', fit.degree=1)
{
    n.obs <- dim(mfcc.data)[1]
    n.loc <- dim(loc.coords)[1]

    if(is.null(dist.mesh)){ 
        graph.weights <- sapply(1:nrow(mesh$edges), function(j) dist(mesh$nodes[mesh$edges[j, ],], method='euclidean'))
        g2 <- add.edges(graph.empty(n), t(mesh$edges), weight=graph.weights )
        dist.mesh <- shortest.paths(g2, mode='all')
    }
    obs.nodes <- rep(NA, n.obs)
    for (i in 1:n.obs){ 
        obs.nodes[i] <- which(mesh$nodes[, 1]==obs.coords[i, 1]& mesh$nodes[, 2]==obs.coords[i, 2])
    }

    loc.nodes <- rep(NA, n.loc)
    for (i in 1:n.loc){ 
        loc.nodes[i] <- which(mesh$nodes[, 1]==loc.coords[i, 1]& mesh$nodes[, 2]==loc.coords[i, 2])
    }

    ctr.cov <- array(NA, c(n.loc, dim(mfcc.data)[2], dim(mfcc.data)[3], dim(mfcc.data)[3]))
    if (is.null(mfcc.center)){
        for (k in 1:n.loc){
            index <- which(mesh$nodes[obs.nodes, 1]==loc.coords[k, 1]& mesh$nodes[obs.nodes, 2]==loc.coords[k, 2])
            ctr.cov[k, ,,] <- cov_process_sqrt(mfcc.data[index, ,], centered=FALSE)
        }
    }else{
        mfcc.centered <- array(NA, c(n.obs, dim(mfcc.data)[2], dim(mfcc.data)[3]))
        for (i in 1:n.obs){
            index <- which(loc.coords[, 1]==mesh$nodes[obs.nodes[i], 1] & loc.coords[, 2]==mesh$nodes[obs.nodes[i], 2])[1]
            mfcc.centered[i, ,] <- mfcc.data[i, ,]-mfcc.center[index, ,]
        }
        for (k in 1:n.loc){
            index <- which(obs.nodes==loc.nodes[k])
            ctr.cov[k, ,,] <- cov_process_sqrt(mfcc.centered[index, ,], centered=TRUE)
        }
    }

    log.cov <- ctr.cov
    for (k in 1:dim(ctr.cov)[1]){
        for (t in 1:dim(ctr.cov)[2]){
            log.cov[k, t,, ] <- sqrtM(ctr.cov[k, t,, ])
        }
    }

    ## set output.indices according to the output
    if( output == 'cross-validation' )
    {
        ##
        cov.cv <- numeric(nrow(loc.coords))
        for( k.ind in 1:nrow(loc.coords) )
        {
            k <- loc.nodes[k.ind]

            if( bandwidth.mode == 'inversedist' )
            {
                lfit <- compute_pointwise_cov_fit_inversedist(log.cov[-k.ind, ,,],
                                                              current.coord=mesh$nodes[k,],
                                                              loc.coords=loc.coords[-k.ind,],
                                                              weight.scaling=rep(1, dim(loc.coords)[1])[-k.ind],
                                                              local.dists=dist.mesh[k, loc.nodes[-k.ind]],  alpha=alpha)
            } else {
                tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode, weight.scaling= n.sd[-k.ind], local.dists=dist.mesh[k, loc.nodes[-k.ind]], fit.degree)

                ## compute pointwise fit
                lfit <- compute_pointwise_cov_fit(tau, log.cov[-k.ind, ,,],
                                                  current.coord=mesh$nodes[k,],
                                                  loc.coords=loc.coords[-k.ind,],
                                                  weight.scaling=rep(1, dim(loc.coords)[1])[-k.ind],
                                                  local.dists=dist.mesh[k, loc.nodes[-k.ind]], fit.degree=fit.degree)

            }

            cov.predict <- lfit$smoothed.logcov

            d <- dim(cov.predict)
            #d.names <- dimnames(obs.cov)
            cov.predict <- t(apply(cov.predict, 1, '%^%', 2))
            #cov.predict <- aperm(cov.predict, c(2, 3,1))
            cov.predict  <- array(cov.predict, d, dimnames=dimnames(cov.predict)[1])  
            rm(d)
            ## store results
            cov.cv[k.ind] <-  dist_cov_mfcc(cov.predict, ctr.cov[k.ind,,,], dist=sqrtdist)^2
        }

        return(cov.cv)

    } else 
    {
        if( is.matrix(output) )
        {
            output.indices <- numeric(nrow(output))
            for (i in 1:dim(output)[1])
                output.indices[i] <- which(mesh$nodes[, 1]==output[i, 1] & mesh$nodes[, 2]==output[i, 2])
        }  else if (output == 'estimation')
        {
            print(paste0("cov estimation"))
            output.indices  <- 1:nrow(mesh$nodes)
        } else if(output == 'prediction')
        {
            print(paste0("*cov prediction "))
            output.indices <- loc.nodes
        }
        else
            stop(sprintf('undefined value or type of \'output\' %s', output))

        n <- length(output.indices)

        obs.cov <- array(NA, c(n, dim(ctr.cov)[-1]), dimnames=list(output.indices=output.indices))
        taus <- numeric(n)
        effective.number.obs <- numeric(n)

        pb <- txtProgressBar(1, 1+n, style=1)

        for (k.ind in seq(along=output.indices)){
            k <- output.indices[k.ind]

            if( bandwidth.mode == 'inversedist' )
            {
                lfit <- compute_pointwise_cov_fit_inversedist(log.cov,
                                                              current.coord=mesh$nodes[k,],
                                                              loc.coords=loc.coords,
                                                              weight.scaling=rep(1, dim(loc.coords)[1]),
                                                              local.dists=dist.mesh[k, loc.nodes], alpha=alpha)
            } else {
                tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode = bandwidth.mode, weight.scaling = n.sd, local.dists=dist.mesh[k, loc.nodes], fit.degree)


                ## compute pointwise fit
                lfit <- compute_pointwise_cov_fit(tau, log.cov,
                                                  current.coord=mesh$nodes[k,],
                                                  loc.coords=loc.coords,
                                                  weight.scaling=rep(1, dim(loc.coords)[1]),
                                                  local.dists=dist.mesh[k, loc.nodes], fit.degree=fit.degree)
                taus[k.ind] <- lfit$tau ## tau might have been modified by compute_pointwise_cov_fit
            }


            ## store results
            obs.cov[k.ind, , ,] <- lfit$smoothed.logcov
            effective.number.obs[k.ind] <- lfit$effective.number.obs ## tau might have been modified by compute_pointwise_cov_fit

            setTxtProgressBar(pb, k.ind)
        }

        d <- dim(obs.cov)
        #d.names <- dimnames(obs.cov)
        obs.cov <- apply(obs.cov, 1:2, '%^%', 2)
        obs.cov <- aperm(obs.cov, c(2, 3,1))
        obs.cov  <- array(obs.cov, d, dimnames=dimnames(obs.cov)[1])  
        rm(d)
        #for (k in 1:dim(obs.cov)[1]){
        #  for (t in 1:dim(obs.cov)[2]){
        #    obs.cov[k, t,, ] <- powerM(obs.cov[k, t,, ],2)
        #  }
        #}

        out <- c(list(cov=obs.cov), mget(c('mesh', 'h', 'NN', 'taus', 'bandwidth.mode', 'fit.degree', 'effective.number.obs')))
        return(invisible(out))
    }
}


## might be slightly faster?
# cov.smooth.faster <-function(mfcc.data, obs.coords, loc.coords, mfcc.center=NULL, h,mesh, n.sd=rep(1, dim(loc.coords)[1]), dist.mesh=NULL, NN=4, output = 'estimation', bandwidth.mode, fit.degree) #bandwidth.mode='location.NN', fit.degree=1)
# {
#     n.obs <- dim(mfcc.data)[1]
#     n.loc <- dim(loc.coords)[1]
# 
#     if(is.null(dist.mesh)){ 
#         graph.weights <- sapply(1:nrow(mesh$edges), function(j) dist(mesh$nodes[mesh$edges[j, ],], method='euclidean'))
#         g2 <- add.edges(graph.empty(n), t(mesh$edges), weight=graph.weights )
#         dist.mesh <- shortest.paths(g2, mode='all')
#     }
#     obs.nodes <- rep(NA, n.obs)
#     for (i in 1:n.obs){ 
#         obs.nodes[i] <- which(mesh$nodes[, 1]==obs.coords[i, 1]& mesh$nodes[, 2]==obs.coords[i, 2])
#     }
# 
#     loc.nodes <- rep(NA, n.loc)
#     for (i in 1:n.loc){ 
#         loc.nodes[i] <- which(mesh$nodes[, 1]==loc.coords[i, 1]& mesh$nodes[, 2]==loc.coords[i, 2])
#     }
# 
#     ctr.cov <- array(NA, c(n.loc, dim(mfcc.data)[2], dim(mfcc.data)[3], dim(mfcc.data)[3]))
#     if (is.null(mfcc.center)){
#         for (k in 1:n.loc){
#             index <- which(mesh$nodes[obs.nodes, 1]==loc.coords[k, 1]& mesh$nodes[obs.nodes, 2]==loc.coords[k, 2])
#             ctr.cov[k, ,,] <- cov_process(mfcc.data[index, ,], center=TRUE)
#         }
#     }else{
#         mfcc2 <- array(NA, c(n.obs, dim(mfcc.data)[2], dim(mfcc.data)[3]))
#         for (i in 1:n.obs){
#             index <- which(loc.coords[, 1]==mesh$nodes[obs.nodes[i], 1] & loc.coords[, 2]==mesh$nodes[obs.nodes[i], 2])[1]
#             mfcc2[i, ,] <- mfcc.data[i, ,]-mfcc.center[index, ,]
#         }
#         for (k in 1:n.loc){
#             index <- which(obs.nodes==loc.nodes[k])
#             ctr.cov[k, ,,] <- cov_process(mfcc2[index, ,], center=FALSE)
#         }
#     }
# 
#     log.cov <- ctr.cov
#     for (k in 1:dim(ctr.cov)[1]){
#         for (t in 1:dim(ctr.cov)[2]){
#             log.cov[k, t,, ] <- sqrtM(ctr.cov[k, t,, ])
#         }
#     }
# 
#     ## set output.indices according to the output
#     if( is.matrix(output) )
#     {
#         output.indices <- numeric(nrow(output))
#         for (i in 1:dim(output)[1])
#             output.indices[i] <- which(mesh$nodes[, 1]==output[i, 1] & mesh$nodes[, 2]==output[i, 2])
#     }  else if (output == 'estimation')
#     {
#         print(paste0("cov estimation"))
#         output.indices  <- 1:nrow(mesh$nodes)
#     } else if(output == 'prediction')
#     {
#         print(paste0("*cov prediction "))
#         output.indices <- loc.nodes
#     }
#     else
#         stop(sprintf('undefined value or type of \'output\' %s', output))
# 
#     n <- length(output.indices)
# 
#     obs.cov <- array(NA, c(n, dim(ctr.cov)[-1]), dimnames=list(output.indices=output.indices))
#     taus <- numeric(n)
# 
#     pb <- txtProgressBar(1, 1+n, style=1)
# 
#     log.cov.mat <- matrix(log.cov, nrow=dim(log.cov)[1])
# 
#     for (k.ind in seq(along=output.indices)){
#         k <- output.indices[k.ind]
# 
#         tau <- compute_pointwise_bandwidth(h, NN, bandwidth.mode = bandwidth.mode, weight.scaling = n.sd, local.dists=dist.mesh[k, loc.nodes], fit.degree)
# 
# 
#         ## compute pointwise fit
#         lfit <- compute_pointwise_cov_fit_fast(tau, log.cov.mat,
#                                       current.coord=mesh$nodes[k,],
#                                       loc.coords,
#                                       weight.scaling=n.sd,
#                                       total.variance=1,
#                                       local.dists=dist.mesh[k, loc.nodes], fit.degree=fit.degree)
# 
#         ## store results
#         obs.cov[k.ind, , ,] <- lfit$betas
#         taus[k.ind] <- lfit$tau ## tau might have been modified by compute_pointwise_cov_fit
# 
#         setTxtProgressBar(pb, k.ind)
#     }
# 
#     d <- dim(obs.cov)
#     #d.names <- dimnames(obs.cov)
#     obs.cov <- apply(obs.cov, 1:2, '%^%', 2)
#     obs.cov <- aperm(obs.cov, c(2, 3,1))
#     obs.cov  <- array(obs.cov, d, dimnames=dimnames(obs.cov)[1])  
#     rm(d)
#     #for (k in 1:dim(obs.cov)[1]){
#     #  for (t in 1:dim(obs.cov)[2]){
#     #    obs.cov[k, t,, ] <- powerM(obs.cov[k, t,, ],2)
#     #  }
#     #}
# 
#     out <- c(list(cov=obs.cov), mget(c('mesh', 'h', 'NN', 'taus', 'bandwidth.mode', 'fit.degree')))
#     return(invisible(out))
# }
# 
# 

#### bandwidth selection
#' @examples
#' 
#' @section Details:
#'
#'
#
  
smooth_cv <-function(mfcc.data, local.bnc, border.nodes, min.angle=NA, what='mean', h.seq, NN.seq, bandwidth.mode='location.NN', fit.degree, max.area=NA)
{

    prepared.data <- prepare_mfccs(mfcc.data, local.bnc, border.nodes, min.angle, max.area)


    # if(what == 'cov.sqrt' & length(NN.seq)>1 ){
    #     NN.seq <- NA
    #     warning('NN.seq not used for cov.sqrt')
    # }
    cv <- prepared.data %$% array(NA, c(dim(loc.coords)[1], length(h.seq), length(NN.seq)), dimnames=list(dimnames(loc.coords)[[1]], paste0('h=', h.seq), paste0('NN=', NN.seq)))
    #cov.cv <- array(NA, c(dim(loc.coords)[1], length(h.seq), length(NN.seq)), dimnames=list(dimnames(loc.coords)[[1]], paste0('h=', h.seq), paste0('NN=', NN.seq)))
    n.obs <- prepared.data %$% dim(mfcc.data)[1]
    n.loc <- prepared.data %$% dim(cov.data)[1]


    for(h.ind in seq(along=h.seq))
    {
        h <- h.seq[h.ind]
        for(NN.ind in seq(along=NN.seq))
        {
            NN <- NN.seq[NN.ind]

                print(sprintf('== Bandwidth cv == h: %.0f %% -- NN: %.0f %% ==', 100*h.ind/length(h.seq), 100*NN.ind/length(NN.seq)))

            ## only compute the cross-validation that is asked for
            cv[, h.ind, NN.ind] <- switch(what,
                   mean = prepared.data %$% mean_smooth(mfcc.data = mfcc.data, cov.data = cov.data, obs.coords = obs.coords, loc.coords = loc.coords, h=h, mesh = mesh, dist.mesh = dist.mesh, output='cross-validation', NN=NN, bandwidth.mode=bandwidth.mode, fit.degree=fit.degree),
                   cov.sqrt =  prepared.data %$% cov_smooth_sqrt(mfcc.data = mfcc.data, obs.coords = obs.coords, loc.coords = loc.coords, mfcc.center = NULL, h=h, mesh=mesh, dist.mesh = dist.mesh, n.sd = n.sd, output='cross-validation', NN=NN, bandwidth.mode=bandwidth.mode, fit.degree=fit.degree),
                   cov.l2 =  prepared.data %$% cov_smooth_l2(mfcc.data = mfcc.data, obs.coords = obs.coords, loc.coords = loc.coords, mfcc.center = NULL, h=h, mesh=mesh, dist.mesh = dist.mesh, n.sd = n.sd, output='cross-validation', NN=NN, bandwidth.mode=bandwidth.mode, fit.degree=fit.degree)
                   )
        }
    }
    out <- c(list(h=h.seq, NN=NN.seq, cv=cv), mget(c('bandwidth.mode', 'fit.degree')))
    class(out) <- 'smooth_cv'
    return(invisible(out))
}

plot.smooth_cv <- function(smooth.cv, type='image', main.title = '', ...)
{

#    ## ask for next plot
#    if (.Device != "null device") {
#        oldask <- grDevices::devAskNewPage(ask = TRUE)
#        if (!oldask) 
#            on.exit(grDevices::devAskNewPage(oldask), add = TRUE)
#    }
#    op <- options(device.ask.default = TRUE)
#    on.exit(options(op), add = TRUE)
    means <- apply(smooth.cv$cv, 2:3, mean)
    sds <- apply(smooth.cv$cv, 2:3, sd)

    single_plot <- function(what, varname, by, main.title, ...) {
        cols <- tim.colors(ncol(what))
        pchs <- 1:(ncol(what))
        op <- par( oma=c(0, 1,1,0), mai=c(.9,.4, .1, .1))
        #ylim <- range(what)
        matplot(smooth.cv$h, what, pch=pchs, type='b',  log='y', lty=1, col=cols, xlab=varname, ...)
        title(main=main.title, outer=TRUE)
#        par( fig=c(0,1,0,1), oma=rep(0,4), mai=rep(0,4),  new=TRUE)
#        plot(0,0, type='n', bty='n', xaxt='n', yaxt='n')
        legend('topright', lty=1, col=cols, pch=pchs,  legend=paste0(by, '=', smooth.cv$NN), xpd=TRUE, inset=c(0,0), horiz=FALSE, ncol=1)
        par(op)
        return(0)
    }
    

    switch(type,
           h = {
               cols <- tim.colors(ncol(means))
               op <- par( oma=c(6, 3,3,3), mai=c(0,.8, .1, .1), mfrow = c(1,2) )
               #ylim <- range(means)
               matplot(smooth.cv$h, means, pch=20, type='b', log='y', lty=1, col=cols,...)
               matplot(smooth.cv$h, sds, pch=20, type='b', log='y', lty=1, col=cols,...)
               title(main=main.title, outer=TRUE)
               #    matlines(smooth.cv$h, means + sds,  type='l', lty=1)
               #    matlines(smooth.cv$h, means - sds,  type='l', lty=1)
               par( fig=c(0,1,0,1), oma=rep(0,4), mai=rep(0,4),  new=TRUE)
               plot(0,0, type='n', bty='n', xaxt='n', yaxt='n')
               legend('bottom', lty=1, col=cols,  legend=paste0('k=', smooth.cv$NN), xpd=TRUE, inset=c(0,0), horiz=FALSE, ncol=5)
               par(op)
               return(0)
           },
           h.mean = single_plot(means, varname='h', by='k', main.title=main.title, ...),
           h.sd = single_plot(sds, varname='h', by='k', main.title=main.title, ...),
           NN = 
           {
               cols <- tim.colors(nrow(means))
               op <- par( oma=c(6, 3,3,3), mai=c(0,.8, .1, .1), mfrow = c(1,2) )
               ylim <- range(means)
               matplot(smooth.cv$NN, t(means), pch=20, type='b', ylim=ylim, log='y', lty=1, col=cols)
               matplot(smooth.cv$NN, t(sds), pch=20, type='b', log='y', lty=1, col=cols)
               #    matlines(smooth.cv$h, means + sds,  type='l', lty=1)
               #    matlines(smooth.cv$h, means - sds,  type='l', lty=1)
               title(main=main.title, outer=TRUE)
               par( fig=c(0,1,0,1), oma=rep(0,4), mai=rep(0,4),  new=TRUE)
               plot(0,0, type='n', bty='n', xaxt='n', yaxt='n')
               legend('bottom', lty=1, col=cols,  legend=paste0('h=', smooth.cv$h), xpd=TRUE, inset=c(0,0), horiz=FALSE, ncol=5)
               par(op)
               return(0)
           },
           image = 
           {
               op <- par(mfrow=c(1,2), mai=c(1,.8,.8,.4), oma=c(0,0,2,0))
               image.plot(x=smooth.cv$h, y= smooth.cv$NN, means, graphics.reset=TRUE)
               image.plot(x=smooth.cv$h, y= smooth.cv$NN, sds, graphics.reset=TRUE)
               par( fig=c(0,1,0,1), oma=rep(2,4), mai=rep(0,4),  new=TRUE)
               plot(0,0, type='n', bty='n', xaxt='n', yaxt='n')
               title(main=main.title, outer=TRUE)
               par(op) 
               return(0)
           }
           )
}



mfcc_smooth_play_old <- function(data.name, h, nn, nc=5)
{

    ## load data
    #print(load(paste0(work.dir, "data/MFCC-aa-vowels-lframe-16-res-300-channel-1.RData")))
    print(load(paste0(work.dir, "data/MFCC-aa-vowels-lframe-80-res-40-channel-1.RData")))

## select mfccs

    print('loading data')
    # load smooth field
    filen <- sprintf("data/smooth_field-%s-%.1f-%i.RData", data.name, h, nn)
    print(filen)
    load(file=filen)
    mesh <- smooth.field$mesh
    mfccs <- smooth.field$mfcc
    ylim <- range(mfccs[,,1:nc])
    # load initial vowel data and compute f0
    if(!exists(paste0(data.name, ".vowels")))
        print(load(paste0(work.dir, "data/MFCC-aa-vowels-lframe-80-res-40-channel-1.RData")))
    #f0 <- colMeans(aast.vowels$f0, na.rm=TRUE)[floor(seq(from=1, to=ncol(aast.vowels$f0), len=dim(smooth.field$mfcc)[2]))]
    f0 <- aast.vowels$f0[2, floor(seq(from=1, to=ncol(aast.vowels$f0), len=dim(smooth.field$mfcc)[2]))]

    main=sprintf('%s, h=%.1f, NN=%d', data.name, h, nn)
    # load  PC proj
    load( file=sprintf("data/smooth_field-%s--PCproj.RData", data.name))

    print('Ready! ==> left-click on the map to hear corresponding sound -- right-click to exit')
    dev.new()
    dev.image <- dev.cur()
    dev.new()
    dev.mfcc <- dev.cur()
    dev.set(dev.image)
    image.plot(pc.mean[[1]][[paste(h), paste(nn)]], main=main, xaxt='n', yaxt='n', bty='n')
    while( length( nodei <-  identify(mesh$nodes, plot=FALSE, n=1)) > 0) 
    {

        dev.set(dev.image)
        image.plot(pc.mean[[1]][[paste(h), paste(nn)]], main=main, xaxt='l', yaxt='l', bty='n')
        points(mesh$nodes[nodei,, drop=FALSE], pch=20)
        print(nodei)
        ## find node
        ## play sound of node

        mfcc_play(list(mfcc=smooth.field$mfcc[nodei,,1:nc], f0=f0), lframe=80)

        dev.set(dev.mfcc)
        #par(mfrow=c(nc,1))
        matplot((mfccs[nodei,,1:nc]), type='o', ylim=ylim)

        dev.set(dev.image)

    }
    dev.off(dev.image)
    dev.off(dev.mfcc)
}


mfcc_smooth_play <- function(smooth.field, nc=5)
{

    mesh <- smooth.field$mesh
    mfccs <- smooth.field$mfcc
    ylim <- range(mfccs[,,1:nc])
    # load initial vowel data and compute f0
    if(!exists(paste0(data.name, ".vowels")))
        print(load(paste0(work.dir, "data/MFCC-aa-vowels-lframe-80-res-40-channel-1.RData")))
    #f0 <- colMeans(aast.vowels$f0, na.rm=TRUE)[floor(seq(from=1, to=ncol(aast.vowels$f0), len=dim(smooth.field$mfcc)[2]))]
    f0 <- aast.vowels$f0[2, floor(seq(from=1, to=ncol(aast.vowels$f0), len=dim(smooth.field$mfcc)[2]))]

    print('Ready! ==> left-click on the map to hear corresponding sound -- right-click to exit')
    dev.new()
    dev.image <- dev.cur()
    dev.new()
    dev.mfcc <- dev.cur()
    dev.set(dev.image)
    im <- plot(smooth.field, what='mean.pc.smooth', res=200)
    while( length( nodei <-  identify(mesh$nodes, plot=FALSE, n=1)) > 0) 
    {
        dev.set(dev.image)
        image.plot(im)
        points(smooth.field[['loc.coords']], pch=20)
        points(mesh$nodes[nodei,, drop=FALSE], pch=1)
        print(nodei)
        ## find node
        ## play sound of node

        mfcc_play(list(mfcc=smooth.field$mfcc[nodei,,1:nc], f0=f0), lframe=120)

        dev.set(dev.mfcc)
        #par(mfrow=c(nc,1))
        matplot((mfccs[nodei,,1:nc]), type='o', ylim=ylim)

        dev.set(dev.image)

    }
    dev.off(dev.image)
    dev.off(dev.mfcc)
}

## compute the pointwise mean fit for the duration
compute_pointwise_mean_fit_duration <- function(tau, duration.local.means, current.coord, loc.coords, weight.fct=gauss.kern, weight.scaling, total.variance, local.dists, fit.degree=1)
{
    #browser()
            #weights <- rep(NA, n.loc)
            #X <- matrix(NA, n.loc, 3)
            X <- cbind(1 , sweep(loc.coords, 2, current.coord))
            X <- as.matrix(X)
    if(fit.degree == 0)
    {
        X <- X[,1, drop=FALSE]
    } else if (fit.degree != 1)
    {
        stop(sprintf('bad value of fit.degree %s in call %s', fit.degree, deparse(sys.call()) ) )
    }

    weights <- weight.scaling*weight.fct(local.dists, tau)
    while( isTRUE(all.equal(0, sum(weights))) ) ## if tau is too small
    {
        tau <- tau*1.1
        tau.modif <- TRUE
        weights <- weight.scaling*weight.fct(local.dists, tau)
    }
    weights <- weights/sum(weights)

    effective.number.obs <- 1/sum(weights^2 / weight.scaling)

    weights <- weights/total.variance

    ## solve the weighted least squares

    ## this line could be faster
    y.tmp <- matrix(duration.local.means, nrow(duration.local.means))
    betas <- lm.wfit( y=y.tmp , x=X , w=weights)$coef
    ## this is a bit slower
    #betas <- lm(Y ~ X-1, weights=weights, data=list(Y=matrix(mfcc.data, dim(mfcc.data)[1] , prod(dim(mfcc.data)[-1])), X=X))$coef
    smoothed.duration  <- betas[1] 
    if(exists('tau.modif')) 
    {
        warning(sprintf('tau modified to %.2f', tau))
        rm(tau.modif)
    }

    return(mget(c('smoothed.duration', 'effective.number.obs', 'tau')))
}

## could make this much faster for the case fit.degree = 0
compute_pointwise_mean_fit <- function(tau, mfcc.data.local.means, current.coord, loc.coords, weight.fct=gauss.kern, weight.scaling, total.variance, local.dists, fit.degree=1)
{
    #browser()
            #weights <- rep(NA, n.loc)
            #X <- matrix(NA, n.loc, 3)
            X <- cbind(1 , sweep(loc.coords, 2, current.coord))
    if(fit.degree == 0)
    {
        X <- X[,1, drop=FALSE]
    } else if (fit.degree != 1)
    {
        stop(sprintf('bad value of fit.degree %s in call %s', fit.degree, deparse(sys.call()) ) )
    }

    weights <- weight.scaling*weight.fct(local.dists, tau)
    while( isTRUE(all.equal(0, sum(weights))) ) ## if tau is too small
    {
        tau <- tau*1.1
        tau.modif <- TRUE
        weights <- weight.scaling*weight.fct(local.dists, tau)
    }
    weights <- weights/sum(weights)

    effective.number.obs <- 1/sum(weights^2 / weight.scaling)

    weights <- weights/total.variance

    ## solve the weighted least squares

    ## this line could be faster
    y.tmp <- matrix(mfcc.data.local.means, dim(mfcc.data.local.means)[1] , prod(dim(mfcc.data.local.means)[-1]))
    betas <- lm.wfit( y=y.tmp , x=X , w=weights)$coef
    ## this is a bit slower
    #betas <- lm(Y ~ X-1, weights=weights, data=list(Y=matrix(mfcc.data, dim(mfcc.data)[1] , prod(dim(mfcc.data)[-1])), X=X))$coef
    smoothed.mfcc <- array(betas[1, ], dim(mfcc.data.local.means)[-1])
    if(exists('tau.modif')) 
    {
        warning(sprintf('tau modified to %.2f', tau))
        rm(tau.modif)
    }

    return(mget(c('smoothed.mfcc', 'effective.number.obs', 'tau')))
}

compute_pointwise_cov_fit <- function(tau, log.cov, current.coord, loc.coords, weight.fct=gauss.kern, weight.scaling, local.dists, fit.degree=1)
{

    #browser()
            #weights <- rep(NA, n.loc)
            #X <- matrix(NA, n.loc, 3)
            X <- cbind(1 , sweep(loc.coords, 2, current.coord))
    if(fit.degree == 0)
    {
        X <- X[,1, drop=FALSE]
    } else if (fit.degree != 1)
        stop(sprintf('bad value of fit.degree %s in call %s', fit.degree, deparse(sys.call()) ) )

    weights <- weight.scaling*weight.fct(local.dists, tau)
    while( isTRUE(all.equal(0, sum(weights))) ) ## if tau is too small
    {
        tau <- tau*1.1
        tau.modif <- TRUE
        weights <- weight.scaling*weight.fct(local.dists, tau)
    }
    weights <- weights/sum(weights)

    effective.number.obs <- 1/sum(weights^2 / weight.scaling)


    ## solve the weighted least squares

    ## this line could be faster
    y.tmp <- matrix(log.cov, dim(log.cov)[1] , prod(dim(log.cov)[-1]))
    betas <- lm.wfit( y=y.tmp , x=X , w=weights)$coef
    ## this is a bit slower
    #betas <- lm(Y ~ X-1, weights=weights, data=list(Y=matrix(mfcc.data, dim(mfcc.data)[1] , prod(dim(mfcc.data)[-1])), X=X))$coef
    smoothed.logcov <- array(betas[1, ], dim(log.cov)[-1])
    if(exists('tau.modif')) 
    {
        warning(sprintf('tau modified to %.2f', tau))
        rm(tau.modif)
    }

    return(mget(c('smoothed.logcov', 'effective.number.obs', 'tau')))
}


compute_pointwise_cov_fit_inversedist <- function(log.cov, current.coord, loc.coords, weight.fct=gauss.kern, weight.scaling, local.dists, alpha)
{

    X <- matrix(1, nrow=nrow(loc.coords))

    weights <- inverse_dist_weights(local.dists, alpha)
    weights <- weights/sum(weights)

    effective.number.obs <- 1/sum(weights^2 / weight.scaling)


    ## solve the weighted least squares

    ## this line could be faster
    y.tmp <- matrix(log.cov, dim(log.cov)[1] , prod(dim(log.cov)[-1]))
    betas <- lm.wfit( y=y.tmp , x=X , w=weights)$coef
    ## this is a bit slower
    #betas <- lm(Y ~ X-1, weights=weights, data=list(Y=matrix(mfcc.data, dim(mfcc.data)[1] , prod(dim(mfcc.data)[-1])), X=X))$coef
    smoothed.logcov <- array(betas[1, ], dim(log.cov)[-1])

    return(mget(c('smoothed.logcov', 'effective.number.obs')))
}

## might be slightly faster?
compute_pointwise_cov_fit_fast <- function(tau, Y, current.coord, loc.coords, weight.fct=gauss.kern, weight.scaling, total.variance, local.dists, fit.degree=1)
{
    #browser()
            #weights <- rep(NA, n.loc)
            #X <- matrix(NA, n.loc, 3)
            X <- cbind(1 , sweep(loc.coords, 2, current.coord))
    if(fit.degree == 0)
    {
        X <- X[,1, drop=FALSE]
    } else if (fit.degree != 1)
        stop(sprintf('bad value of fit.degree %s in call %s', fit.degree, deparse(sys.call()) ) )

            weights <- weight.scaling*weight.fct(local.dists, tau)
    while( isTRUE(all.equal(0, sum(weights))) ) ## if tau is too small
    {
        tau <- tau*1.1
        tau.modif <- TRUE
        weights <- weight.scaling*weight.fct(local.dists, tau)
    }
    weights <- weights/sum(weights)

    effective.number.obs <- 1/sum(weights^2 / weight.scaling)

    weights <- weights/total.variance

    ## solve the weighted least squares

    ## this line could be faster
    betas <- lm.wfit( y=Y , x=X , w=weights)$coef[1,]
    ## this is a bit slower
    #betas <- lm(Y ~ X-1, weights=weights, data=list(Y=matrix(mfcc.data, dim(mfcc.data)[1] , prod(dim(mfcc.data)[-1])), X=X))$coef
    if(exists('tau.modif')) 
    {
        warning(sprintf('tau modified to %.2f', tau))
        rm(tau.modif)
    }

    return(mget(c('betas', 'effective.number.obs', 'tau')))
}


## bandwidth.mode = c('effective.NN', 'fixed', 'location.NN')
compute_pointwise_bandwidth <- function(h, NN, bandwidth.mode, weight.scaling, local.dists, fit.degree)
{

## only keep positive local distances ## not used, so that the bandwidth is not discontinuous at observation points
#    to.keep <- (local.dists > 0)
#    local.dists <- local.dists[to.keep] 
    #weight.scaling <- weight.scaling[to.keep]

    n.loc <- length(local.dists)
    if( bandwidth.mode == 'fixed' ){
        tau <- h
    } else if (bandwidth.mode == 'effective.NN') ## choose tau with respect to distance to at least number of observations; this could be improved upon by taking a fractional part to get exactly a certain number of observations within distance tau
    {
        nmin <- ifelse(fit.degree == 0, 1, 2) # (set it to distance to nmin-th NN at least)
        od <- order(local.dists)
        ## find the distance to at least NN obs 
        taui <- min(n.loc, max(nmin, which(cumsum(weight.scaling[od]) >= NN)[1]), na.rm=TRUE)
        tau <- h*local.dists[od[taui]]
    } else if (bandwidth.mode == 'location.NN') ## choose tau with respect to 
    {
        tau <- h*sort(local.dists)[NN]
    } else
        stop(sprintf('unknown bandwidth mode %s', bandwidth.mode))
    return(tau)
}

gauss.kern <- function(x, tau)
{
    if(tau > 0)
     return((1/(2*pi*tau^2))*exp(-0.5*(x^2)/tau^2))
else
    return( ifelse(x==0, 1, 0) )
}


plot.smooth_field <- function(smooth.field, what='mean.pc',  pci=1, rci=2, from=NULL, direction=NULL, res=100, lambda=1e-10, cex=.5, cov.std=FALSE, ...)
{
#    if( !exists('uk.border.nodes.lowres') )
#        load(file=sprintf('%sdata/uk-border-lowres.RData', work.dir)) 

    switch(what,
           mean.proj = {
               if( is.null(direction) ) stop('direction must be specified')
               if( !(all.equal(dim(direction) , dim(smooth.field$mfcc)[-1]) == TRUE) ) stop('direction must be of same dimension as MFCCs')
               ## compute scores of mean field
               meanfield.scores <- rep(NA, dim(smooth.field$mesh$nodes)[1])
               for (k in 1:dim(smooth.field$mesh$nodes)[1]){
                   meanfield.scores[k]<- sum(direction*(smooth.field$mfcc[k, ,]))
               }

               smooth.field %$% {
              mesh_values_plot(mesh, mesh[['nodes']], meanfield.scores, loc.coords=loc.coords, main=sprintf('Mean Field projected on specific direction'))
               }
           },
           mean.proj.smooth = {
               if( is.null(direction) ) stop('direction must be specified')
               if( !(all.equal(dim(direction) , dim(smooth.field$mfcc)[-1]) == TRUE) ) stop('direction must be of same dimension as MFCCs')
               ## compute scores of mean field
               meanfield.scores <- rep(NA, dim(smooth.field$mesh$nodes)[1])
               for (k in 1:dim(smooth.field$mesh$nodes)[1]){
                   meanfield.scores[k]<- sum(direction*(smooth.field$mfcc[k, ,]))
               }

               smooth.field %$% {
              mesh_smooth_plot(mesh, mesh[['nodes']], meanfield.scores, lambda=lambda, res=res, loc.coords=loc.coords, border=border.nodes, main=sprintf('Mean Field projected on specific direction'), ...)
               }
           },
           mean.pc = {
               ## compute PCs
               pc <- 
                   smooth.field %$% pca %$% 
                   {
                       if( dim(prcomps)[1] < pci ) stop(sprintf('pc %d not
                                                                available in
                                                                %s', pci,
                                                                deparse(sys.call())))
                       prcomps[pci,,]
                   }

               ## compute scores of mean field
               meanfield.scores <- rep(NA, dim(smooth.field$mesh$nodes)[1])
               for (k in 1:dim(smooth.field$mesh$nodes)[1]){
                   meanfield.scores[k]<- sum(pc*(smooth.field$mfcc[k, ,]))
               }

               smooth.field %$% {
              mesh_values_plot(mesh, mesh[['nodes']], meanfield.scores, loc.coords=loc.coords, main=sprintf('Mean Field projected on PC %d', pci), ...)
               }
           },
           mean.pc.smooth = {
               ## compute PCs
               pc <- 
                   smooth.field %$% pca %$% 
                   {
                       if( dim(prcomps)[1] < pci ) stop(sprintf('pc %d not
                                                                available in
                                                                %s', pci,
                                                                deparse(sys.call())))
                       prcomps[pci,,]
                   }

               ## compute scores of mean field
               meanfield.scores <- rep(NA, dim(smooth.field$mesh$nodes)[1])
               for (k in 1:dim(smooth.field$mesh$nodes)[1]){
                   meanfield.scores[k]<- sum(pc*(smooth.field$mfcc[k, ,]))
               }

               smooth.field %$% {
              mesh_smooth_plot(mesh, mesh[['nodes']], meanfield.scores, lambda=lambda, res=res, loc.coords=NA, border=border.nodes, main=sprintf('Mean Field projected on PC %d', pci), cex=cex, ...)
               }
           },
           mean.effective.nobs = {
               smooth.field %$% mesh_values_plot(mesh, mesh$nodes, loc.coords=loc.coords, values=mean.effective.nobs, cex=cex, main=sprintf('Effective number of observation'), ...)
           },
           mean.effective.nobs.smooth = {
               smooth.field %$% mesh_smooth_plot(mesh, mesh$nodes, what=mean.effective.nobs, loc.coords=loc.coords, cex=cex, main=sprintf('Effective number of observation'), lambda=lambda, res=res, border.nodes=border.nodes, ...)
           },
        cov.effective.nobs = {
               smooth.field %$% mesh_values_plot(mesh, mesh$nodes, loc.coords=loc.coords, values=cov.effective.nobs, cex=cex, main=sprintf('Effective number of observation'), ...)
           },
        cov.effective.nobs.smooth = {
               smooth.field %$% mesh_smooth_plot(mesh, mesh$nodes, what=cov.effective.nobs, loc.coords=loc.coords, cex=cex, main=sprintf('Effective number of observation'), lambda=lambda, res=res, border.nodes=border.nodes, ...)
           },
           mean.bandwidth = {
               smooth.field %$% {
                   mesh_values_plot(mesh, mesh$nodes, loc.coords=loc.coords, values=mean.taus, cex=cex, main='Bandwidth used', ...)
               }
           },
           mean.bandwidth.smooth = {
               smooth.field %$% 
               {
              mesh_smooth_plot(mesh, mesh[['nodes']], mean.taus, lambda=lambda, res=res, loc.coords=loc.coords, border=border.nodes, main='Bandwidth used', ...) %>% {return(invisible(.))}
               }
           },
           cov.bandwidth = {
               smooth.field %$% {
                   mesh_values_plot(mesh, mesh$nodes, loc.coords=loc.coords, values=cov.taus, cex=cex, main='Bandwidth used', ...)
               }
           },
           cov.bandwidth.smooth = {
               smooth.field %$% 
               {
              mesh_smooth_plot(mesh, mesh[['nodes']], cov.taus, lambda=lambda, res=res, loc.coords=loc.coords, border=border.nodes, main='Bandwidth used', ...) %>% {return(invisible(.))}
               }
           },
           residuals = {
               mesh_values_plot(smooth.field[['mesh']], where=smooth.field[['loc.coords']], loc.coords=NA, values=residuals.smooth_field(smooth.field), cex=cex, main='Residuals of the Fit', ...)
           },
           residuals.smooth = {
               smooth.field %$% {
              mesh_smooth_plot(mesh, where=loc.coords, residuals.smooth_field(smooth.field), lambda=lambda, res=res, loc.coords=loc.coords, border=border.nodes, main='Residuals of the Fit', ...)
               }
           },
           mean.energy = {
               mesh_values_plot(smooth.field[['mesh']], where=smooth.field[[ c('mesh', 'nodes') ]], loc.coords=NA, values=apply(smooth.field[['mfcc']], 1, Matrix::norm, type='F'), cex=cex, main='Energy (HS Norm) of the mean field', ...)
           },
           mean.energy.diagnostic = {
        plot(sort(apply(smooth.field[['mfcc']], 1, Matrix::norm, type='F')), pch='.', main='Sorted value of Energy (HS norm) of the mean field', ylab='', ...)
           },
           mean.energy.smooth = {
               smooth.field %$% {
              mesh_smooth_plot(mesh, mesh[['nodes']], apply(smooth.field[['mfcc']], 1, Matrix::norm, type='F'), lambda=lambda, res=res, loc.coords=loc.coords, border=border.nodes, main='Energy (HS Norm) of the mean field', ...)
               }
           },
           cov.trace = {
               mesh_values_plot(smooth.field[['mesh']], where=smooth.field[[ c('mesh', 'nodes') ]], 
                                loc.coords=NA,
                                values= ( apply(smooth.field[['cov']], c(1, 3:4), mean) %>%
                                         apply(., 1, function(x) sum(diag(x)) ) ) ,
                                cex=cex, main='Total Variation', ...)
           },
           cov.trace.smooth = {
               mesh_smooth_plot(smooth.field[['mesh']], where=smooth.field[[ c('mesh', 'nodes') ]], 
                                what= ( apply(smooth.field[['cov']], c(1, 3:4), mean) %>%
                                         apply(., 1, function(x) sum(diag(x)) ) ) ,
                                cex=cex, main='Total Variation', lambda=lambda, res=res, loc.coords=smooth.field[['loc.coords']], border=border.nodes, ...)
           },
        cov.cor.smooth = {
          mesh_smooth_plot(smooth.field[['mesh']], where=smooth.field[[ c('mesh', 'nodes') ]], 
                           what= ( apply(smooth.field[['cov']], c(1, 3:4), mean) %>%
                                     apply(., 1, function(x) x[pci,rci]/(sqrt(x[pci,pci]*x[rci,rci])) ) ) ,
                           cex=cex, main=paste0('d-Correlation MFCC',pci,'-MFCC',rci), lambda=lambda, res=res, loc.coords=smooth.field[['loc.coords']], border=border.nodes, ...)
        },
           cov.opnorm = {
               val <- ( apply(smooth.field[['cov']], c(1:2), base::norm, type='2') %>% rowMeans )
               mesh_values_plot(smooth.field[['mesh']], where=smooth.field[[ c('mesh', 'nodes') ]], 
                                loc.coords=smooth.field[['loc.coords']],
                                values= val  ,
                                cex=cex, main='Operator Norm of Covariance', ...)
               return(invisible(val))
           },
           cov.opnorm.smooth = {
               val <- ( apply(smooth.field[['cov']], c(1:2), base::norm, type='2') %>% rowMeans )
               mesh_smooth_plot(smooth.field[['mesh']], where=smooth.field[[ c('mesh', 'nodes') ]], 
                                what= val, 
                                cex=cex, main='Operator Norm of Covariance', lambda=lambda, res=res, loc.coords=smooth.field[['loc.coords']], border=border.nodes, ...)
               return(invisible(val))
           },
           dist.from.cov  = {
               if( is.null(from) ) stop('from must be specified')

               index <- smooth.field %$% which(apply(mesh$nodes, 1, identical, loc.coords[from,]))
               scores <- rep(NA, dim(smooth.field$mesh$nodes)[1])
#               cov1 <- smooth.field$cov[index, ,,]
#               cov1.tr <- apply(cov1, 1, function(x) sum(diag(x)))
#               ## renormalization of the trace of the d-covariance
#               cov1 <- cov1 %>% apply(2:3, function(y) y/cov1.tr) 
               for (k in 1:dim(smooth.field$mesh$nodes)[1]){
                   ## renormalization of the trace of the d-covariance
                   #cov2 <- smooth.field$cov[k, ,,]
#                   cov2.tr <- apply(cov2, 1, function(x) sum(diag(x)))
#                   cov2 <- cov2 %>% apply(2:3, function(y) y/cov2.tr) 
                   scores[k] <- dist_cov_mfcc(smooth.field$cov[index, ,,], smooth.field$cov[k, ,,], dist=sqrtdist, std=cov.std)
               }
               smooth.field %$% {
                   mesh_smooth_plot(mesh, mesh[['nodes']], scores, lambda=lambda, res=res, loc.coords=NA, border=border.nodes, main=sprintf('Covariance Distance from %s', dimnames(loc.coords)[[c(1, from)]]), cex=cex, ...)
               }
           })
}

mesh_values_plot <- function(mesh, where, loc.coords=NA, values, ncols=72, cex=.5, ...)
{
    #browser()
    if( sd(values) == 0){
        range.loc <- values[1]*c(1-1e-1, 1+1e-1)
    } else  {
        range.loc <- range(values)
    }
    #op <- par(oma=c(0,0,0,4))
    mai.tmp <- par('mai')
    #mai.tmp[4] <- 1
    op <- par(mai=mai.tmp)
    cols <- tim.colors(ncols, alpha=.8*255)[cut(values, breaks = ncols)]
    plot(mesh, pch='.', xaxt='n', yaxt='n', lwd=.3, ...)
    points(where, pch=20, col=cols, cex=cex)
    if(is.matrix(loc.coords)) points(loc.coords, pch=17)
    par(op)
    image.plot(legend.only=TRUE, nlevels=ncols, zlim=range.loc)
}

mesh_smooth_plot <- function(mesh, where, what, lambda, res, loc.coords=NA, border=border, bty='n', xaxt='n', yaxt='n', xlab='', ylab='', ...)
{
#    if( !exists('uk.border.nodes.lowres') )
#        load(file=sprintf('%sdata/uk-border-lowres.RData', work.dir)) 
    basis <- create.FEM.basis(mesh)
    scores_fit <- smooth.FEM.basis(where, what, basis, lambda=lambda, GCV=FALSE)

    return(invisible(FEM.image(FEM=scores_fit$fit.FEM, res=res, data.locations=loc.coords, border=border, doplot=TRUE, bty=bty, xaxt=xaxt, yaxt=yaxt, xlab=xlab, ylab=ylab, ...)))
}

residuals.smooth_field <- function(smooth.field)
{
              smooth.field %$% {
                      sapply(1:nrow(loc.coords), function(i) {
                      loc.node.tmp <- which(apply(mesh$nodes, 1, identical, loc.coords[i,]))
                      fit.tmp <- mfcc[loc.node.tmp,, ]
                      obs.nodes.tmp <- which(apply(obs.coords, 1, identical, loc.coords[i,]))
                      residuals.tmp <- mfcc.data[obs.nodes.tmp,,]
                      sweep(residuals.tmp, 2:3, fit.tmp) %>% 
                            apply(., 1, Matrix::norm, type='F') %>% 
                            mean 
                  }) %>% {return(invisible(.))}
              }
}


inverse_dist_weights <- function(local.dists, alpha)
{
    zero.dist <- local.dists == 0
    if( any(zero.dist) ) {
        ans <-  rep(0, length(local.dists))
        ans[zero.dist] <- 1  
        return(ans)
    } else {
        return( ( 1/local.dists )^alpha  )
    }
}
